#!/usr/bin/env pytest

# Unit test for debug_plugin's URL encoding

from ornithology import *
from sys import platform
import os
import classad2
import subprocess
import signal
from pathlib import Path

DEFAULT_HOSTNAME = "default.test.hostname"

# TestName : (URL, Expected Ad, Expected Exit Code, input ad(filename, contents), ad producing script(filename, contents))
ENCODING_TEST_CASES = {
    "success" : (
        "debug://success",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : True,
            "TransferFileBytes" : 0,
            "TransferTotalBytes" : 0,
            "ConnectionTimeSeconds" : 0,
        },
        0,
        None,
        None,
    ),
    "error" : (
        "debug://error",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "TransferFileBytes" : 0,
            "TransferTotalBytes" : 0,
            "ConnectionTimeSeconds" : 0,
        },
        1,
        None,
        None,
    ),
    "inline_general" : (
        "debug://success/general[test_var=True;TransferFileBytes=100;TransferTotalBytes=100;ConnectionTimeSeconds=18;]",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : True,
            "TransferFileBytes" : 100,
            "TransferTotalBytes" : 100,
            "ConnectionTimeSeconds" : 18,
            "test_var" : True,
        },
        0,
        None,
        None,
    ),
    "file_general" : (
        "debug://error/general~[sample.ad]/foo/bar",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "TransferFileBytes" : 42,
            "TransferTotalBytes" : 13,
            "ConnectionTimeSeconds" : 1000,
            "TestVar1" : False,
            "TestVar2" : 2.03,
            "TestVar3" : "Hello World",
        },
        1,
        (
            "sample.ad",
"""TransferFileBytes = 42
TransferTotalBytes = 13
ConnectionTimeSeconds = 1000
TestVar1 = False
TestVar2 = 2.03
TestVar3 = "Hello World"
"""
        ),
        None,
    ),
    "script_general" : (
        "debug://error/general#[./custom.sh]",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "TransferFileBytes" : 10,
            "TransferTotalBytes" : 9,
            "TransferStartTime" : 1234,
            "TransferEndTime" : 1235,
            "ConnectionTimeSeconds" : 1,
            "TestVar1" : True,
        },
        1,
        None,
        (
            "custom.sh",
"""#!/bin/bash
echo 'TestVar1 = True'
echo 'TransferFileBytes = 10'
echo 'TransferTotalBytes = 9'
echo 'ConnectionTimeSeconds = 1'
echo 'TransferStartTime = 1234'
echo 'TransferEndTime = 1235'
"""
        ),
    ),
    "ignore_trailing_info" : (
        "debug://success/general[test=False;]ignore-this-garbage-with-no-error/general[test=True;]more-garbage/foo",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : True,
            "TransferFileBytes" : 0,
            "TransferTotalBytes" : 0,
            "ConnectionTimeSeconds" : 0,
            "test" : True,
        },
        0,
        None,
        None,
    ),
    "universal_space_replacement" : (
        'debug://success/general[TestSpaces="Hello%20World"]',
        {
            "TestSpaces" : "Hello World",
        },
        0,
        None,
        None,
    ),
    "inline_special_space_replacement" : (
        'debug://success/general[TestVar="A+B+C\\+D"]',
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : True,
            "TransferFileBytes" : 0,
            "TransferTotalBytes" : 0,
            "ConnectionTimeSeconds" : 0,
            "TestVar" : "A B C+D",
        },
        0,
        None,
        None,
    ),
    "script_special_space_replacement" : (
        "debug://success/general#[./space.sh::3::2.3]",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : True,
            "TransferFileBytes" : 0,
            "TransferTotalBytes" : 0,
            "ConnectionTimeSeconds" : 0,
            "TestVar1" : 3,
            "TestVar2" : 2.3,
        },
        0,
        None,
        (
            "space.sh",
"""#!/bin/bash
echo "TestVar1 = $1"
echo "TestVar2 = $2"
"""
        ),
    ),
    "error_parameter_default" : (
        "debug://error/parameter/foo/bar",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "TransferFileBytes" : 0,
            "TransferTotalBytes" : 0,
            "ConnectionTimeSeconds" : 0,
            "TransferErrorData" : [
                {
                    "ErrorType" : "Parameter",
                    "ErrorCode": -1,
                    "ErrorString": "Invalid plugin parameters",
                    "PluginVersion": "1.1.0",
                    "PluginLaunched": True,
                },
            ],
            "TransferError" : "URL transfer encoded to fail",
        },
        1,
        None,
        None,
    ),
    "error_parameter_custom" : (
        'debug://error/parameter[ErrorString="Custom%20failure%20occurred!";ErrorCode=19;PluginLaunched=False;DeveloperData=[TestVar1=False;TestVar2=2.45;];]/foo/bar/baz',
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "TransferFileBytes" : 0,
            "TransferTotalBytes" : 0,
            "ConnectionTimeSeconds" : 0,
            "TransferErrorData" : [
                {
                    "ErrorType" : "Parameter",
                    "ErrorCode": 19,
                    "ErrorString": "Custom failure occurred!",
                    "PluginVersion": "1.1.0",
                    "PluginLaunched": False,
                    "DeveloperData" : {
                        "TestVar1" : False,
                        "TestVar2" : 2.45,
                    }
                },
            ],
            "TransferError" : "URL transfer encoded to fail",
        },
        1,
        None,
        None,
    ),
    "error_resolution_default" : (
        "debug://error/resolution/foo/bar",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "TransferFileBytes" : 0,
            "TransferTotalBytes" : 0,
            "ConnectionTimeSeconds" : 0,
            "TransferErrorData" : [
                {
                    "ErrorType" : "Resolution",
                    "ErrorCode": -2,
                    "ErrorString": "Failed to resolve server name to contact",
                    "FailedName": DEFAULT_HOSTNAME,
                    "FailureType": "Definitive",
                },
            ],
            "TransferError" : "URL transfer encoded to fail",
        },
        1,
        None,
        None,
    ),
    "error_resolution_custom" : (
        "debug://error/resolution~[resolution.ad]/foo/bar",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "TransferFileBytes" : 0,
            "TransferTotalBytes" : 0,
            "ConnectionTimeSeconds" : 0,
            "TransferErrorData" : [
                {
                    "ErrorType" : "Resolution",
                    "ErrorCode": 32,
                    "ErrorString": "Failed because Cole told me to...",
                    "FailedName": "chtc.cs.cthulu",
                    "FailureType": "PostContact",
                    "DeveloperData" : {
                        "TestVar1" : True,
                        "TestVar2" : 69,
                    }
                },
            ],
            "TransferError" : "URL transfer encoded to fail",
        },
        1,
        (
            "resolution.ad",
"""ErrorCode = 32
ErrorString = "Failed because Cole told me to..."
FailedName = "chtc.cs.cthulu"
FailureType = "PostContact"
DeveloperData = [TestVar1 = True;TestVar2 = 69;]
""",
        ),
        None,
    ),
    "error_contact_default" : (
        "debug://error/contact/foo/bar",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "TransferFileBytes" : 0,
            "TransferTotalBytes" : 0,
            "ConnectionTimeSeconds" : 0,
            "TransferErrorData" : [
                {
                    "ErrorType" : "Contact",
                    "ErrorCode": -3,
                    "ErrorString": "Failed to contact server",
                    "FailedServer": DEFAULT_HOSTNAME,
                },
            ],
            "TransferError" : "URL transfer encoded to fail",
        },
        1,
        None,
        None,
    ),
    "error_contact_custom" : (
        "debug://error/contact#[./contact-failure.py]/foo/bar",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "TransferFileBytes" : 0,
            "TransferTotalBytes" : 0,
            "ConnectionTimeSeconds" : 0,
            "TransferErrorData" : [
                {
                    "ErrorType" : "Contact",
                    "ErrorCode": 2,
                    "ErrorString": "This is not a drill",
                    "FailedServer": "test.url.parse.server",
                    "CustomVar" : 123.45,
                },
            ],
            "TransferError" : "URL transfer encoded to fail",
        },
        1,
        None,
        (
            "contact-failure.py",
"""#!/usr/bin/env python3
print("ErrorCode = 2")
print('ErrorString = "This is not a drill"')
print('FailedServer = "test.url.parse.server"')
print("CustomVar = 123.45")
""",
        ),
    ),
    "error_authorization_default" : (
        "debug://error/authorization/foo/bar",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "TransferFileBytes" : 0,
            "TransferTotalBytes" : 0,
            "ConnectionTimeSeconds" : 0,
            "TransferErrorData" : [
                {
                    "ErrorType" : "Authorization",
                    "ErrorCode": -4,
                    "ErrorString": "Failed authorization with server",
                    "FailedServer": DEFAULT_HOSTNAME,
                    "FailureType": "Authorization",
                    "ShouldRefresh": True,
                },
            ],
            "TransferError" : "URL transfer encoded to fail",
        },
        1,
        None,
        None,
    ),
    "error_authorization_custom" : (
        'debug://error/authorization[ErrorCode=23;ErrorString="Untrustworthy%20server";FailureType="Authorization"]/foo/bar',
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "TransferFileBytes" : 0,
            "TransferTotalBytes" : 0,
            "ConnectionTimeSeconds" : 0,
            "TransferErrorData" : [
                {
                    "ErrorType" : "Authorization",
                    "ErrorCode": 23,
                    "ErrorString": "Untrustworthy server",
                    "FailedServer": DEFAULT_HOSTNAME,
                    "FailureType": "Authorization",
                    "ShouldRefresh": True,
                },
            ],
            "TransferError" : "URL transfer encoded to fail",
        },
        1,
        None,
        None,
    ),
    "error_specification_default" : (
        "debug://error/specification/foo/bar",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "TransferFileBytes" : 0,
            "TransferTotalBytes" : 0,
            "ConnectionTimeSeconds" : 0,
            "TransferErrorData" : [
                {
                    "ErrorType" : "Specification",
                    "ErrorCode": -5,
                    "ErrorString": "Specified file DNE",
                    "FailedServer": DEFAULT_HOSTNAME,
                },
            ],
            "TransferError" : "URL transfer encoded to fail",
        },
        1,
        None,
        None,
    ),
    "error_specification_custom" : (
        "debug://error/specification~[specification.failed.ad]/foo/bar",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "TransferFileBytes" : 0,
            "TransferTotalBytes" : 0,
            "ConnectionTimeSeconds" : 0,
            "TransferErrorData" : [
                {
                    "ErrorType" : "Specification",
                    "ErrorCode": -5,
                    "ErrorString": "Specified file DNE",
                    "FailedServer": "random.host.in.world",
                    "DeveloperData" : {
                        "Reason" : "Unspecified host",
                    }
                },
            ],
            "TransferError" : "URL transfer encoded to fail",
        },
        1,
        (
            "specification.failed.ad",
"""FailedServer = "random.host.in.world"
DeveloperData = [Reason="Unspecified host";]
""",
        ),
        None,
    ),
    "error_transfer_default" : (
        "debug://error/transfer/foo/bar",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "TransferFileBytes" : 0,
            "TransferTotalBytes" : 0,
            "ConnectionTimeSeconds" : 0,
            "TransferErrorData" : [
                {
                    "ErrorType" : "Transfer",
                    "ErrorCode": -6,
                    "ErrorString": "Failed to transfer file",
                    "FailedServer": DEFAULT_HOSTNAME,
                    "FailureType": "TooSlow",
                },
            ],
            "TransferError" : "URL transfer encoded to fail",
        },
        1,
        None,
        None,
    ),
    "error_transfer_custom" : (
        'debug://error/transfer[FailureType="NoSpace";ErrorCode=47;DeveloperData=[TestVar1=14;TestVar2=9.09;];ErrorString="Ran%20out%20of%20space"]/foo/bar',
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "TransferFileBytes" : 0,
            "TransferTotalBytes" : 0,
            "ConnectionTimeSeconds" : 0,
            "TransferErrorData" : [
                {
                    "ErrorType" : "Transfer",
                    "ErrorCode": 47,
                    "ErrorString": "Ran out of space",
                    "FailedServer": DEFAULT_HOSTNAME,
                    "FailureType": "NoSpace",
                    "DeveloperData" : {
                        "TestVar1" : 14,
                        "TestVar2" : 9.09,
                    }
                },
            ],
            "TransferError" : "URL transfer encoded to fail",
        },
        1,
        None,
        None,
    ),
    "multiple_failed_attempts" : (
        "debug://error/contact/transfer/contact/authorization/foo/bar/baz",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "TransferFileBytes" : 0,
            "TransferTotalBytes" : 0,
            "ConnectionTimeSeconds" : 0,
            "TransferErrorData" : [
                {
                    "ErrorType" : "Contact",
                    "ErrorCode": -3,
                    "ErrorString": "Failed to contact server",
                    "FailedServer": DEFAULT_HOSTNAME,
                },
                {
                    "ErrorType" : "Transfer",
                    "ErrorCode": -6,
                    "ErrorString": "Failed to transfer file",
                    "FailedServer": DEFAULT_HOSTNAME,
                    "FailureType": "TooSlow",
                },
                {
                    "ErrorType" : "Contact",
                    "ErrorCode": -3,
                    "ErrorString": "Failed to contact server",
                    "FailedServer": DEFAULT_HOSTNAME,
                },
                {
                    "ErrorType" : "Authorization",
                    "ErrorCode": -4,
                    "ErrorString": "Failed authorization with server",
                    "FailedServer": DEFAULT_HOSTNAME,
                    "FailureType": "Authorization",
                    "ShouldRefresh": True,
                },
            ],
            "TransferError" : "URL transfer encoded to fail",
        },
        1,
        None,
        None,
    ),
    "complex_example" : (
        'debug://error/general[GenVar="Testing";TransferFileBytes=1000;TransferTotalBytes=200;]/CoNtAcT[ErrorCode=2;FailedServer="big.dumb.server";DeveloperData=[AttemptedContacts=5;]]/Authorization#[./gen_ad.sh]/foo/bar/transfer~[complex.ad]/baz',
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "TransferFileBytes" : 1000,
            "TransferTotalBytes" : 200,
            "ConnectionTimeSeconds" : 0,
            "GenVar" : "Testing",
            "TransferErrorData" : [
                {
                    "ErrorType" : "Contact",
                    "ErrorCode": 2,
                    "ErrorString": "Failed to contact server",
                    "FailedServer": "big.dumb.server",
                    "DeveloperData" : {
                        "AttemptedContacts" : 5,
                    }
                },
                {
                    "ErrorType" : "Authorization",
                    "ErrorCode": 13,
                    "ErrorString": "Miron doesnt trust this host",
                    "FailedServer": "slurm.super.host",
                    "FailureType": "Authorization",
                    "ShouldRefresh": False,
                    "IgnoreHostInFuture" : True,
                },
                {
                    "ErrorType" : "Transfer",
                    "ErrorCode": 13,
                    "ErrorString": "Failed to get data from pelican director",
                    "FailedServer": "smart.good.server",
                    "FailureType": "TimedOut",
                    "DeveloperData" : {
                        "TimeoutSeconds" : 60,
                        "Attempts" : 3,
                        "ContactedDirector" : True,
                    }
                },
            ],
            "TransferError" : "URL transfer encoded to fail",
        },
        1,
        (
            "complex.ad",
"""ErrorCode = 13
ErrorString = "Failed to get data from pelican director"
FailedServer = "smart.good.server"
FailureType = "TimedOut"
DeveloperData = [TimeoutSeconds=60;Attempts=3;ContactedDirector=True;]
""",
        ),
        (
            "gen_ad.sh",
"""#!/bin/bash
echo 'ErrorCode = 13'
echo 'ErrorString = "Miron doesnt trust this host"'
echo 'FailedServer = "slurm.super.host"'
echo 'ShouldRefresh = False'
echo 'IgnoreHostInFuture = True'
"""
        ),
    ),
    "success_with_failed_attempts" : (
        "debug://success/contact/transfer/foo/bar/baz",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : True,
            "TransferFileBytes" : 0,
            "TransferTotalBytes" : 0,
            "ConnectionTimeSeconds" : 0,
            "TransferErrorData" : [
                {
                    "ErrorType" : "Contact",
                    "ErrorCode": -3,
                    "ErrorString": "Failed to contact server",
                    "FailedServer": DEFAULT_HOSTNAME,
                },
                {
                    "ErrorType" : "Transfer",
                    "ErrorCode": -6,
                    "ErrorString": "Failed to transfer file",
                    "FailedServer": DEFAULT_HOSTNAME,
                    "FailureType": "TooSlow",
                },
            ],
            "TransferError" : "URL transfer encoded to fail",
        },
        0,
        None,
        None,
    ),
    "bad_url_file_dne" : (
        "debug://error/general~[DNE.txt]/foo/contact/baz",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "IsParseFailure" : True,
        },
        124,
        None,
        None,
    ),
    "bad_url_bad_inline_encoding" : (
        "debug://success/general[foo;]/contact",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "IsParseFailure" : True,
        },
        124,
        None,
        None,
    ),
    "bad_url_missing_bracket" : (
        "debug://success/contact[filename/foo",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "IsParseFailure" : True,
            "TransferError" : "DebugUrlParseError: Invalid URL has open bracket with no close.",
        },
        124,
        None,
        None,
    ),
    "bad_url_provide_non_executable" : (
        "debug://error/general#[./not_a_script]",
        {
            "TransferProtocol" : "debug",
            "TransferSuccess" : False,
            "IsParseFailure" : True,
        },
        124,
        ("not_a_script", "Derp"),
        None,
    ),
}

FIRST_TEST_CASE = True
DEBUG_AD_FILE = "result_ads.txt"

@action(params={name: name for name in ENCODING_TEST_CASES})
def test_url_encoding(request, path_to_debug_plugin):
    global FIRST_TEST_CASE
    url, expected_ad, expected_exit, file_info, script_info = ENCODING_TEST_CASES[request.param]

    if FIRST_TEST_CASE:
        FIRST_TEST_CASE = False
        if os.path.exists(DEBUG_AD_FILE):
            os.remove(DEBUG_AD_FILE)

    if file_info is not None:
        f_name, f_contents = file_info
        Path(f_name).write_text(f_contents)

    if script_info is not None:
        name, contents = script_info
        Path(name).write_text(contents)
        os.chmod(name, 0o755)

    p = subprocess.run([path_to_debug_plugin, "-test", url], stdout=subprocess.PIPE)
    p_ad = classad2.parseOne(p.stdout.rstrip().decode())

    if request.param[:7] == "bad_url":
        assert "TransferErrorData" not in p_ad

    with open(DEBUG_AD_FILE, "a") as f:
        f.write(f"{request.param}: {url} result ad:\n")
        f.write(f"{p_ad}\n###########################\n")

    return (expected_ad, expected_exit, p_ad, p.returncode)


@action
def test_ad_encoding_delete(path_to_debug_plugin):
    delete = ["TestVar1","TestVar2"]
    keys = ",".join(delete)
    p = subprocess.run(
        [path_to_debug_plugin, "-test", f"debug://success/general[TestVar1=True;TestVar2=13;TestVar3=2.1;]/delete[{keys}]"],
        stdout=subprocess.PIPE
    )
    assert p.returncode == 0
    p_ad = classad2.parseOne(p.stdout.rstrip().decode())
    return (p_ad, delete)


# bash signal exit codes = 128 + val
EXIT_TEST_CASES = {
    "exit_failure" : ("debug://exit/18", 18),
    "exit_success" : ("debug://exit/0", 0),
    "exit_trailing_info" : ("debug://exit/42/foo/bar/baz", 42),
    "exit_negative" : ("debug://exit/-64", 64),
    "exit_out_of_range" : ("debug://exit/9999", 123),
    "default_signal" : ("debug://signal", 0 - signal.SIGTERM),
    "sig_hangup" : ("debug://signal/SIGHUP", 0 - signal.SIGHUP),
    #"sig_interrupt" : ("debug://signal/SIGINT", 0 - signal.SIGINT),  # This exits 1 on some Linux OS for some reason
    "sig_kill" : ("debug://signal/SIGKILL", 0 - signal.SIGKILL),
    "sig_terminate" : ("debug://signal/SIGTERM", 0 - signal.SIGTERM),
}

if platform != "darwin":
    # These test cases raise an annoying report on MacOs everytime
    EXIT_TEST_CASES.update({
        "sig_abort" : ("debug://signal/SIGABRT", 0 - signal.SIGABRT),
        "sig_segmentation" : ("debug://signal/SIGSEGV", 0 - signal.SIGSEGV),
    })

@action(params={name: name for name in EXIT_TEST_CASES})
def test_plugin_exits(request, path_to_debug_plugin):
    url, code = EXIT_TEST_CASES[request.param]
    p = subprocess.run(
        [path_to_debug_plugin, "-test", url],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    return (code, p.returncode)


@action
def test_plugin_sleep(path_to_debug_plugin):
    try:
        p = subprocess.run(
            [path_to_debug_plugin, "-test", "debug://sleep/10"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            timeout=1,
        )
        return False
    except subprocess.TimeoutExpired:
        return True


def compare_ads(expected, result):
    for key, val in expected.items():
        assert key in result
        if type(val) is dict:
            compare_ads(val, result[key])
        elif type(val) is list:
            for i in range(len(val)):
                compare_ads(val[i], result[key][i])
        else:
            assert val == result[key]

class TestDebugPlugin:
    def test_ad_encoding(self, test_url_encoding):
        expected_ad, expected_exit, ad, code = test_url_encoding
        assert expected_exit == code
        compare_ads(expected_ad, ad)

    def test_encoding_delete(self, test_ad_encoding_delete):
        ad, keys = test_ad_encoding_delete
        for key in keys:
            assert key not in ad.keys()

    def test_specified_exits(self, test_plugin_exits):
        e_code, ret_code = test_plugin_exits
        assert e_code == ret_code

    def test_sleep(self, test_plugin_sleep):
        assert test_plugin_sleep == True



#!/usr/bin/env pytest

import os
import subprocess
from pathlib import Path

import classad2

from ornithology import (
    action,
    Condor,
)

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@action
def the_config_file(test_dir):
    the_config_file = (test_dir / "test_warning.config")
    the_config_file.write_text("""
        DEFAULT_CHECKPOINT_DESTINATION_PREFIX = example://test/dir
        use feature:DefaultCheckpointDestination
    """)
    return the_config_file


TEST_CASES = {
    "no_checkpoint_exit_code": {
        "job": """
            Owner = "example-owner-string"
        """,
        "transformed_job": {
            "SuccessCheckpointExitCode": None,
            "CheckpointDestination": None,
            "CheckpointStorage": None,
        }
    },
    "only_checkpoint_exit_code": {
        "job": """
            Owner = "example-owner-string"
            SuccessCheckpointExitCode = 85
        """,
        "transformed_job": {
            "SuccessCheckpointExitCode": 85,
            "CheckpointDestination": "example://test/dir/example-owner-string",
            "checkpoint_storage": None,
        }
    },
    "checkpoint_storage_default": {
        "job": """
            Owner = "example-owner-string"
            SuccessCheckpointExitCode = 85
            checkpoint_storage = "default"
        """,
        "transformed_job": {
            "SuccessCheckpointExitCode": 85,
            "CheckpointDestination": "example://test/dir/example-owner-string",
            "checkpoint_storage": "default",
        }
    },
    "checkpoint_storage_spool": {
        "job": """
            Owner = "example-owner-string"
            SuccessCheckpointExitCode = 85
            checkpoint_storage = "spool"
        """,
        "transformed_job": {
            "SuccessCheckpointExitCode": 85,
            "CheckpointDestination": None,
            "checkpoint_storage": "spool",
        }
    },
    "checkpoint_storage_custom": {
        "job": """
            Owner = "example-owner-string"
            SuccessCheckpointExitCode = 85
            checkpoint_storage = "custom://myplace"
        """,
        "transformed_job": {
            "SuccessCheckpointExitCode": 85,
            "CheckpointDestination": "custom://myplace",
            "checkpoint_storage": "custom://myplace",
        }
    }
}


@action(params={name: name for name in TEST_CASES})
def the_test_case(request):
    return (request.param, TEST_CASES[request.param])


@action
def the_test_name(the_test_case):
    return the_test_case[0]


@action
def the_test_dict(the_test_case):
    return the_test_case[1]


@action
def the_test_output(test_dir, the_test_name, the_test_dict, the_config_file):
    ad_file = (test_dir / the_test_name)
    ad_file.write_text(the_test_dict["job"])

    p = subprocess.run(
        ["condor_transform_ads", "-jobtransform", "CheckpointStorage", str(ad_file)],
            timeout=60,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            env={** os.environ, "CONDOR_CONFIG": str(the_config_file),}
    )
    logger.debug(p.stderr)
    return p.stdout


class TestDefaultCheckpointDestination:

    def test_warning(self, test_dir):
        config_file = (test_dir / "test_warning.config")
        config_file.write_text("""
            use feature:DefaultCheckpointDestination
        """)

        p = subprocess.run(
            [ "condor_config_val", "use", "feature:DefaultCheckpointDestination" ],
            timeout=60,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            env={** os.environ, "CONDOR_CONFIG": str(config_file),}
        )
        assert p.stderr == " Warning : You must set DEFAULT_CHECKPOINT_DESTINATION_PREFIX or DEFAULT_CHECKPOINT_DESTINATION to use feature:DefaultCheckpointDestination.\n"


    def test_job_transform(self, the_test_output, the_test_dict):
        ad = classad2.parseNext(the_test_output, classad2.ParserType.Old)

        for attribute in the_test_dict["transformed_job"].keys():
            value = ad.get(attribute)
            assert value == the_test_dict["transformed_job"][attribute]




#!/usr/bin/env pytest

#   test_dagman_inline_shell_submit.py
#
#   This test multiple functionalities in regards to DAGMan
#   inline submit descriptions. This includs:
#       1. Works with shell condor_submit
#       2. JOB sub commands work correctly (DIR, DONE, NOOP)
#       3. Support for lists of jobs (queue N)
#       4. Support for '{}' or '@=' enclosing operators
#       5. Support for prepend of job VARS
#       6. Inline works for FINAL, PROVISIONER, & SERVICE nodes
#
#   Author: Cole Bollig - 2024/09/11

from ornithology import *
import htcondor2
import os
from shutil import rmtree
import re
from time import sleep

SUBDIR = "test_subdir"
JOB_LOG = "check.log"
D_CLUSTER = -1

TEST_CASES = {
    "SHELL_SUBMIT" : {
        "contents" : """
# Node A: Queue 1 && Success
JOB A {{
    executable = {sleep_exe}
    arguments  = 0
}}

# Node B: Should be marked as DONE else DAG fails
JOB B @=EndB
    executable = {exit_exe}
    arguments  = 1
@EndB DONE

# Node C: Should be NOOP or else DAG fails waiting
JOB C @=C-DESC
    executable = {sleep_exe}
    arguments  = 360
@C-DESC NOOP

# Node D: Queue 2 && Success && run in sub-directory
JOB D @=blah
    executable = {sleep_exe}
    arguments  = 0
    log = {log}
    queue 2
@blah DIR {dir}""",
        "SubmitMethod" : 0,
        "TotalNodes" : 4,
        "SubmittedJobs" : 3,
    },

    "PREPEND_VARS" : {
        "contents" : """
JOB CUSTOM_VAR_NODE @=prepend-vars-test
    executable = {sleep_exe}
    arguments  = 0
    if defined CUSTOM_VAR
        My.HasCustomVar = True
    else
        My.HasCustomVar = False
    endif
@prepend-vars-test

VARS CUSTOM_VAR_NODE PREPEND CUSTOM_VAR="True" """,
        "SubmitMethod" : 1,
        "TotalNodes" : 1,
        "SubmittedJobs" : 1,
    },

    "SPECIAL_NODES" : {
        "contents" : """
PROVISIONER P {{
    # Ensure provisioner node is around long enough for DAGMan to read
    # go ahead from ProvisionerState ClassAd attr (2)
    executable = {sleep_exe}
    arguments  = 600
    My.ProvisionerState = 2
    My.RemoveTag = True
}}

# >:(   Post script so that provisioner removal doesn't fail dag
SCRIPT POST P {exit_exe} 0

SERVICE S @=my-service-node
    executable = {exit_exe}
    arguments  = 0
    queue 3
@my-service-node

# Job to remove provisioner job so we don't wait forever
JOB A @=desc
    executable = {rm_exe}
    universe   = local
    getenv     = PATH,_CONDOR*,CONDOR_CONFIG
    output     = job.debug
    error      = job.debug
@desc

FINAL F @=final-node
    executable = {exit_exe}
    arguments  = 0
@final-node
        """,
        "SubmitMethod" : 0,
        "TotalNodes" : 3, # Service nodes don't count towards node count
        "SubmittedJobs" : 6,
    }
}

#==================================================================
@standup
def condor(test_dir):
    with Condor(local_dir=test_dir / "condor", config={
            "DAGMAN_REMOVE_TEMP_SUBMIT_FILES" : False}) as condor:
        yield condor

#------------------------------------------------------------------
@action
def write_exit_script(test_dir):
    script = test_dir / "exit.sh"
    with open(script, "w") as f:
        f.write("#!/bin/bash\nexit $1\n")
        os.chmod(script, 0o755)
        return script

#------------------------------------------------------------------
@action
def write_remove_script(test_dir):
    script = test_dir / "removal.sh"
    with open(script, "w") as f:
        f.write("#!/bin/bash\ncondor_rm -const 'RemoveTag==True'\nexit 0\n")
        os.chmod(script, 0o755)
        return script

#------------------------------------------------------------------
@action
def run_dags(condor, test_dir, path_to_sleep, write_exit_script, write_remove_script):
    if os.path.exists(SUBDIR):
        rmtree(SUBDIR)
    os.mkdir(SUBDIR)

    test_info = dict()

    fmt = {
        "sleep_exe" : path_to_sleep,
        "exit_exe" : write_exit_script,
        "rm_exe" : write_remove_script,
        "log" : JOB_LOG,
        "dir" : SUBDIR,
    }

    for test, details in TEST_CASES.items():
        dag_file = test + ".dag"
        with open(dag_file, "w") as f:
            f.write(details["contents"].format(**fmt))
        dag = htcondor2.Submit.from_dag(dag_file, {"SubmitMethod" : details["SubmitMethod"]})
        handle = condor.submit(dag)
        test_info[test] = (handle, details["TotalNodes"], details["SubmittedJobs"])

    yield test_info

#------------------------------------------------------------------
@action(params={name: name for name in TEST_CASES})
def test_info(request, run_dags):
    return (request.param, run_dags[request.param][0], run_dags[request.param][1], run_dags[request.param][2])

@action
def test_name(test_info):
    return test_info[0]

@action
def test_dag(test_info):
    return test_info[1]

@action
def dag_wait(test_dag):
    assert test_dag.wait(condition=ClusterState.all_complete, timeout=300)
    return test_dag

@action
def dag_num_nodes(test_info):
    return test_info[2]

@action
def dag_num_jobs(test_info):
    return test_info[3]

#------------------------------------------------------------------
@action
def test_get_history(condor, dag_wait, dag_num_nodes, dag_num_jobs):
    ads = list()
    const = f'(ClusterId=={dag_wait.clusterid} || DAGManJobId=={dag_wait.clusterid})'
    proj = ["DAGNodeName", "DAG_NodesTotal", "ClusterId", "HasCustomVar", "ExitCode"]

    with condor.use_config():
        # Attempt to get all job records from history archive
        attempt = 0
        while True:
            ads = htcondor2.Schedd().history(constraint=const, projection=proj)
            attempt += 1

            if len(ads) > dag_num_jobs:
                # We expect a record for each submitted job (dag_num_jobs) plus
                # one for the schedule uni. job. Thus, only break out if we have
                # more records than the submitted count.
                # Note: Later check will verify the correct number of records
                #       (greater than or less than) and contents
                break
            elif attempt >= 5:
                # At this point something probably went wrong so return records
                # for likely failure
                break
            else:
                # Sleep for a little bit before attempting query once more
                sleep(5)

    return (ads, dag_num_nodes, dag_num_jobs)

#==================================================================
class TestDAGManInlineDesc:
    def test_check_job_history(self, test_get_history):
        global D_CLUSTER
        num_node_jobs = 0
        num_dag_jobs = 0
        for ad in test_get_history[0]:
            if "DAGNodeName" not in ad:
                num_dag_jobs += 1
                assert ad.get("DAG_NodesTotal") == test_get_history[1]
                assert ad["ExitCode"] == 0
            else:
                num_node_jobs += 1
                if ad["DAGNodeName"] == "CUSTOM_VAR_NODE":
                    assert ad.get("HasCustomVar") == True
                else:
                    assert ad.get("HasCustomVar") == None
                    D_CLUSTER = ad["ClusterId"] if ad["DAGNodeName"] == "D" else D_CLUSTER

        assert num_node_jobs == test_get_history[2]
        assert num_dag_jobs == 1

    def test_check_temp_files(self):
        inline_files = set()
        for path in [".", SUBDIR]:
            inline_files.update([f for f in os.listdir(path) if len(f) > 5 and f[-5:] == ".temp"])

        for f in inline_files:
            assert re.match("^[ADFSP]{1}-inline.[0-9]+.temp", f)

    def test_check_multi_queue_log(self):
        expected_procs = set([0, 1, 2])
        log = os.path.join(SUBDIR, JOB_LOG)
        assert os.path.exists(log)
        jel = htcondor2.JobEventLog(log)
        for event in jel.events(0):
            assert event.cluster == D_CLUSTER
            assert event.proc in expected_procs




import sys
import enum


class DrainType(enum.IntEnum):
    """
    An enumeration of draining policies that can be set to a *condor_startd*.

    The following explanations are reminders; the
    `actual policies <http://htcondor.readthedocs.io/en/latest/admin-manual/ep-policy-configuration.html#machine-states>`_
    are way more complicated.

    .. attribute:: Graceful

        The default; jobs are allowed to continue running until they reach
        their promised `max job retirement time <http://htcondor.readthedocs.io/en/latest/admin-manual/configuration-macros.html#MAXJOBRETIREMENTTIME>`_.  To
        reduce idle time, all jobs may continue to run until the most distant
        such deadline, at which point the draining policy effectively becomes
        the following.

    .. attribute:: Quick

        Jobs are immediately soft-killed (generally meaning sent a ``SIGTERM``).
        The startd then waits a certain amount of time to allow jobs to shut
        down cleanly before the draining policy becomes the following.

    .. attribute:: Fast

        Jobs are immediately hard-killed (generally meaning sent a ``SIGKILL``).
    """

    Graceful = 0
    Quick    = 10
    Fast     = 20



#!/usr/bin/pytest

import pytest

import os
import time
import subprocess
from pathlib import Path

#
# Construct a synthetic directory, `syndicate`, underneath $(LOCK), which has
# (multiple) examples of each of the five kinds of files; some should be
# deleted and some shouldn't be.  Then run condor_preen.  Then (the actual
# tests) check to make sure the files that should have been spared still
# remain, and that the files that should have been removed were.
#

REMOVED = [
    [ "oldkeyfile",
      "oldremovelock.rm_1", "oldremovelock.rm_2",
      "oldkeyfile.message",
      "oldkeyfile.1234", "oldkeyfile.5678",
      ".oldkeyfile.432",
    ],
]

PRESERVED = [
    [ "newkeyfile",
      "newremovelock.rm_1", "newremovelock.rm_2",
      "newkeyfile.message",
      "newkeyfile.1234", "newkeyfile.5678",
      ".newkeyfile.432",
    ],
]


@pytest.fixture(params=REMOVED)
def removed(request):
    return request.param


@pytest.fixture(params=PRESERVED)
def preserved(request):
    return request.param


@pytest.fixture
def populated(test_dir, removed, preserved):
    syndicate = test_dir / "syndicate"
    syndicate.mkdir(parents=True, exist_ok=True)
    os.chdir(syndicate.as_posix())

    for file in removed:
        Path(file).touch()
        then = int(time.time()) - 301
        os.utime(file, (then, then))
    for file in preserved:
        Path(file).touch()

    return Path(syndicate.as_posix())


@pytest.fixture
def preened(test_dir, populated):
    env = {
        ** os.environ,
        '_CONDOR_LOG': test_dir.as_posix(),
        '_CONDOR_LOCK': test_dir.as_posix(),
        '_CONDOR_TOOL_DEBUG': 'D_TEST D_FULLDEBUG D_SUB_SECOND D_CATEGORY',
    }

    rv = subprocess.run(
        ['condor_preen', '-d'],
        env=env,
        timeout=60,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
    )
    print(rv.stdout)

    # We could write more complicated logic in condor_preen to try and
    # remove files which depend on other files being missed or expired
    # in a single pass, but there's really no need; this isn't a
    # correctness requirement.
    rv = subprocess.run(
        ['condor_preen', '-d'],
        env=env,
        timeout=60,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
    )
    print(rv.stdout)

    rv = subprocess.run(
        ['condor_preen', '-d'],
        env=env,
        timeout=60,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
    )
    print(rv.stdout)

    return rv.returncode == 0


class TestCIFPreen:

    def test_files(self, preened, populated, removed, preserved):
        assert(preened)

        os.chdir(populated.as_posix())
        for file in preserved:
            assert Path(file).exists()
        for file in removed:
            assert not Path(file).exists()



#!/usr/bin/env pytest

#Configuration for personal condor. The test is sensitive to slot re-use
# so we want the startd to have enough slots to start all jobs at once
#testreq: personal
"""<<CONDOR_TESTREQ_CONFIG
    NUM_CPUS = 20
    CLAIM_WORKLIFE = 0
"""
#endtestreq

import pytest
from ornithology import *
import classad2
import time

#--------------------------------------------------------------------------
# Test Case Structure:
# TestName : (
#     Submit Description Dictionary,
#     Ornithology ClusterState Wait Condition,
#     (
#          Expected Plugin Return Ad Details,
#          Plugin Transfer Ad Type (input/output),
#          Expected Hold Code,
#          Expected Hold SubCode (Actual plugin exit code that we bit shift left by 8),
#     )
# )

TEST_CASES = {
    # Check Input transfer plugin exit code
    "INPUT_FAIL_CODE" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_input_files" : "debug://exit/32",
        },
        ClusterState.all_held,
        (
            {},
            None,
            13,
            32 << 8,
            None,
            None,
        )
    ),
    # Check output transfer plugin exit code
    "OUTPUT_FAIL_CODE" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_output_files" : "foo",
            "transfer_output_remaps" : '"foo=debug://exit/67"',
        },
        ClusterState.all_held,
        (
            {},
            None,
            12,
            67 << 8,
            None,
            None,
        )
    ),
    # Check input transfer plugin receives signal
    "SIGNAL_INPUT" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_input_files" : "debug://signal/SIGKILL",
        },
        ClusterState.all_held,
        (
            {},
            None,
            13,
            9,
            { "ExitBySignal": True, "ExitSignal": 9 },
            "input",
        )
    ),
    # Checkout output transfer plugin receives signal
    "SIGNAL_OUTPUT" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_output_files" : "foo",
            "transfer_output_remaps" : '"foo=debug://signal/SIGKILL"',
        },
        # Presently, the job goes idle under the assumption that it's not a
        # job-specific problem causing the plug-in to die on a signal.
        #
        # This probably doesn't work right now, since it's also the initial state.
        ClusterState.all_idle,
        (
            {},
            None,
            12,
            9,
            { "ExitBySignal": True, "ExitSignal": 9 },
            "output",
        )
    ),
    # Check input transfer plugin failure with parameter case
    "PARAMETER_FAILURE_INPUT" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_input_files" : 'debug://error/parameter[ErrorString="Missing%20host";ErrorCode=19;PluginLaunched=False;DeveloperData=[ConnectionTime=0.07;];]',
        },
        ClusterState.all_held,
        (
            {
                "TransferProtocol" : "debug",
                "TransferSuccess" : False,
                "TransferFileBytes" : 0,
                "TransferTotalBytes" : 0,
                #"ConnectionTimeSeconds" : 0,
                "TransferErrorData" : [
                    {
                        "ErrorType" : "Parameter",
                        "ErrorCode": 19,
                        "ErrorString": "Missing host",
                        "PluginVersion": "1.1.0",
                        "PluginLaunched": False,
                        "DeveloperData" : {
                            "ConnectionTime" : 0.07,
                        }
                    },
                ],
                "TransferError" : "URL transfer encoded to fail",
            },
            "input",
            13,
            1 << 8,
            None,
            None,
        )
    ),
    # Check output transfer plugin failure with parameter case
    "PARAMETER_FAILURE_OUTPUT" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_output_files" : "foo",
            "transfer_output_remaps" : '"foo=debug://error/parameter~[$(parameter_output)]"',
        },
        ClusterState.all_held,
        (
            {
                "TransferProtocol" : "debug",
                "TransferSuccess" : False,
                "TransferFileBytes" : 0,
                "TransferTotalBytes" : 0,
                #"ConnectionTimeSeconds" : 0,
                "TransferErrorData" : [
                    {
                        "ErrorType" : "Parameter",
                        "ErrorCode": 32,
                        "ErrorString": "Invalid host name provided",
                        "PluginVersion": "1.1.0",
                        "PluginLaunched": True,
                        "DeveloperData" : {
                            "ConnectionTime" : 98.3,
                        }
                    },
                ],
                "TransferError" : "URL transfer encoded to fail",
            },
            "output",
            12,
            1 << 8,
            None,
            None,
        )
    ),
    # Check input transfer plugin failure with resolution case
    "RESOLUTION_FAILURE_INPUT" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_input_files" : ",".join([
                'debug://error/resolution[ErrorString="Insufficient%20Sacrifice";ErrorCode=91;FailedName="cthulu.old.god";FailureType="PostContact"]',
                "debug://exit/91",
            ])
        },
        ClusterState.all_held,
        (
            {
                "TransferProtocol" : "debug",
                "TransferSuccess" : False,
                "TransferFileBytes" : 0,
                "TransferTotalBytes" : 0,
                #"ConnectionTimeSeconds" : 0,
                "TransferErrorData" : [
                    {
                        "ErrorType" : "Resolution",
                        "ErrorCode" : 91,
                        "ErrorString" : "Insufficient Sacrifice",
                        "FailedName" : "cthulu.old.god",
                        "FailureType" : "PostContact",
                    },
                ],
                "TransferError" : "URL transfer encoded to fail",
            },
            "input",
            13,
            91 << 8,
            None,
            None,
        ),
    ),
    # Check output transfer plugin failure with resolution case
    "RESOLUTION_FAILURE_OUTPUT" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_output_files" : "foo",
            "transfer_output_remaps" : '"foo=debug://error/resolution~[$(resolution_output)]"',
        },
        ClusterState.all_held,
        (
            {
                "TransferProtocol" : "debug",
                "TransferSuccess" : False,
                "TransferFileBytes" : 0,
                "TransferTotalBytes" : 0,
                #"ConnectionTimeSeconds" : 0,
                "TransferErrorData" : [
                    {
                        "ErrorType" : "Resolution",
                        "ErrorCode" : 404,
                        "ErrorString" : "Failed to locate host",
                        "FailedName" : "perfect.gaming.host",
                        "FailureType" : "Definitive",
                        "DeveloperData" : {
                            "ConnectionAttempts" : 10,
                            "Trolled" : True,
                        }
                    },
                ],
                "TransferError" : "URL transfer encoded to fail",
            },
            "output",
            12,
            1 << 8,
            None,
            None,
        ),
    ),
    # Check input transfer plugin failure with contact case
    "CONTACT_FAILURE_INPUT" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_input_files" : 'debug://error/contact[ErrorString="Lost%20connection";ErrorCode=4;FailedServer="flaky.server.com";IntermediateServerErrorType="Connection";IntermediateServer="flaky.server.com";]',
        },
        ClusterState.all_held,
        (
            {
                "TransferProtocol" : "debug",
                "TransferSuccess" : False,
                "TransferFileBytes" : 0,
                "TransferTotalBytes" : 0,
                #"ConnectionTimeSeconds" : 0,
                "TransferErrorData" : [
                    {
                        "ErrorType" : "Contact",
                        "ErrorCode" : 4,
                        "ErrorString" : "Lost connection",
                        "FailedServer" : "flaky.server.com",
                        "IntermediateServerErrorType" : "Connection",
                        "IntermediateServer" : "flaky.server.com",
                    },
                ],
                "TransferError" : "URL transfer encoded to fail",
            },
            "input",
            13,
            1 << 8,
            None,
            None,
        ),
    ),
    # Check output transfer plugin failure with contact case
    "CONTACT_FAILURE_OUTPUT" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_output_files" : "foo",
            "transfer_output_remaps" : '"foo=debug://error/contact~[$(contact_output)]"',
        },
        ClusterState.all_held,
        (
            {
                "TransferProtocol" : "debug",
                "TransferSuccess" : False,
                "TransferFileBytes" : 0,
                "TransferTotalBytes" : 0,
                #"ConnectionTimeSeconds" : 0,
                "TransferErrorData" : [
                    {
                        "ErrorType" : "Contact",
                        "ErrorCode" : 12,
                        "ErrorString" : "Failed to contact server",
                        "FailedServer" : "national.secrets.gov",
                    },
                ],
                "TransferError" : "URL transfer encoded to fail",
            },
            "output",
            12,
            1 << 8,
            None,
            None,
        ),
    ),
    # Check input transfer plugin failure with authorization case
    "AUTHORIZATION_FAILURE_INPUT" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_input_files" : 'debug://error/authorization[ErrorCode=78;ErrorString="Not%20authorized";FailedServer="moate.chtc.wisc.edu";FailureType="Authorization";ShouldRefresh=False;]'
        },
        ClusterState.all_held,
        (
            {
                "TransferProtocol" : "debug",
                "TransferSuccess" : False,
                "TransferFileBytes" : 0,
                "TransferTotalBytes" : 0,
                #"ConnectionTimeSeconds" : 0,
                "TransferErrorData" : [
                    {
                        "ErrorType" : "Authorization",
                        "ErrorCode" : 78,
                        "ErrorString" : "Not authorized",
                        "FailedServer" : "moate.chtc.wisc.edu",
                        "FailureType" : "Authorization",
                        "ShouldRefresh" : False,
                    },
                ],
                "TransferError" : "URL transfer encoded to fail",
            },
            "input",
            13,
            1 << 8,
            None,
            None,
        ),
    ),
    # Check output transfer plugin failure with authorization case
    "AUTHORIZATION_FAILURE_OUTPUT" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_output_files" : "foo",
            "transfer_output_remaps" : '"foo=debug://error/authorization~[$(authorization_output)]"',
        },
        ClusterState.all_held,
        (
            {
                "TransferProtocol" : "debug",
                "TransferSuccess" : False,
                "TransferFileBytes" : 0,
                "TransferTotalBytes" : 0,
                #"ConnectionTimeSeconds" : 0,
                "TransferErrorData" : [
                    {
                        "ErrorType" : "Authorization",
                        "ErrorCode" : 52,
                        "ErrorString" : "Detected bad cookies",
                        "FailedServer" : "cookie.monster.org",
                        "FailureType" : "Authentication",
                        "ShouldRefresh" : True,
                        "DeveloperData" : {
                            "Note" : "Clear cookie cache",
                        },
                    },
                ],
                "TransferError" : "URL transfer encoded to fail",
            },
            "output",
            12,
            1 << 8,
            None,
            None,
        ),
    ),
    # Check input transfer plugin failure with specification case
    "SPECIFICATION_FAILURE_INPUT" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_input_files" : 'debug://error/specification[ErrorCode=404;ErrorString="File%20foo%20DNE";FailedServer="todd.chtc.wisc.edu";]'
        },
        ClusterState.all_held,
        (
            {
                "TransferProtocol" : "debug",
                "TransferSuccess" : False,
                "TransferFileBytes" : 0,
                "TransferTotalBytes" : 0,
                #"ConnectionTimeSeconds" : 0,
                "TransferErrorData" : [
                    {
                        "ErrorType" : "Specification",
                        "ErrorCode" : 404,
                        "ErrorString" : "File foo DNE",
                        "FailedServer" : "todd.chtc.wisc.edu",
                    },
                ],
                "TransferError" : "URL transfer encoded to fail",
            },
            "input",
            13,
            1 << 8,
            None,
            None,
        ),
    ),
    # Check output transfer plugin failure with specification case
    "SPECIFICATION_FAILURE_OUTPUT" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_output_files" : "foo",
            "transfer_output_remaps" : '"foo=debug://error/specification~[$(specification_output)]"',
        },
        ClusterState.all_held,
        (
            {
                "TransferProtocol" : "debug",
                "TransferSuccess" : False,
                "TransferFileBytes" : 0,
                "TransferTotalBytes" : 0,
                #"ConnectionTimeSeconds" : 0,
                "TransferErrorData" : [
                    {
                        "ErrorType" : "Specification",
                        "ErrorCode" : 405,
                        "ErrorString" : "Failed to create file foo",
                        "FailedServer" : "todd.chtc.wisc.edu",
                    },
                ],
                "TransferError" : "URL transfer encoded to fail",
            },
            "output",
            12,
            1 << 8,
            None,
            None,
        ),
    ),
    # Check input transfer plugin failure with transfer case
    "TRANSFER_FAILURE_INPUT" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_input_files" : 'debug://error/transfer[ErrorCode=9;ErrorString="Out%20of%20space";FailedServer="cole.chtc.wisc.edu";FailureType="NoSpace";]'
        },
        ClusterState.all_held,
        (
            {
                "TransferProtocol" : "debug",
                "TransferSuccess" : False,
                "TransferFileBytes" : 0,
                "TransferTotalBytes" : 0,
                #"ConnectionTimeSeconds" : 0,
                "TransferErrorData" : [
                    {
                        "ErrorType" : "Transfer",
                        "ErrorCode" : 9,
                        "ErrorString" : "Out of space",
                        "FailedServer" : "cole.chtc.wisc.edu",
                        "FailureType" : "NoSpace",
                    },
                ],
                "TransferError" : "URL transfer encoded to fail",
            },
            "input",
            13,
            1 << 8,
            None,
            None,
        ),
    ),
    # Check output transfer plugin failure with transfer case
    "TRANSFER_FAILURE_OUTPUT" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_output_files" : "foo",
            "transfer_output_remaps" : '"foo=debug://error/transfer~[$(transfer_failure_output)]"',
        },
        ClusterState.all_held,
        (
            {
                "TransferProtocol" : "debug",
                "TransferSuccess" : False,
                "TransferFileBytes" : 0,
                "TransferTotalBytes" : 0,
                #"ConnectionTimeSeconds" : 0,
                "TransferErrorData" : [
                    {
                        "ErrorType" : "Transfer",
                        "ErrorCode" : 10,
                        "ErrorString" : "Transfer rate fell below threshold: 100 B/s",
                        "FailedServer" : "tj.winblows.net",
                        "FailureType" : "TooSlow",
                    },
                ],
                "TransferError" : "URL transfer encoded to fail",
            },
            "output",
            12,
            1 << 8,
            None,
            None,
        ),
    ),
    # Check input transfer plugin failure with multiple failed attempts
    "MULTI_ATTEMPT_FAILURE_INPUT" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_input_files" : ",".join([
                "/".join([
                    "debug://error",
                    'contact[ErrorCode=19;ErrorString="TCP%20handshake%20failed";FailedServer="pelican.origin-2.org"]',
                    'authorization[ErrorCode=404;ErrorString="Token%20BAR%20failed%20auth";FailedServer="pelican.origin-2.org";FailureType="Authentication";ShouldRefresh=False]',
                    'transfer[ErrorCode=5;ErrorString="Below%20threshold(5KB/s)";FailedServer="pelican.origin-2.org";FailureType="TooSlow"]',
                    'transfer[ErrorCode=6;ErrorString="Timer%20expired";FailedServer="pelican.origin-2.org";FailureType="TimedOut"]',
                    'transfer[ErrorCode=7;ErrorString="Exceeded%20pull%20rates";FailedServer="pelican.origin-2.org";FailureType="Quota"]',
                ]),
                "debug://exit/72",
            ])
        },
        ClusterState.all_held,
        (
            {
                "TransferProtocol" : "debug",
                "TransferSuccess" : False,
                "TransferFileBytes" : 0,
                "TransferTotalBytes" : 0,
                #"ConnectionTimeSeconds" : 0,
                "TransferErrorData" : [
                    {
                        "ErrorType" : "Contact",
                        "ErrorCode" : 19,
                        "ErrorString" : "TCP handshake failed",
                        "FailedServer" : "pelican.origin-2.org",
                    },
                    {
                        "ErrorType" : "Authorization",
                        "ErrorCode" : 404,
                        "FailedServer" : "pelican.origin-2.org",
                        "FailureType" : "Authentication",
                        "ShouldRefresh" : False,
                    },
                    {
                        "ErrorType" : "Transfer",
                        "ErrorCode" : 5,
                        "ErrorString" : "Below threshold(5KB/s)",
                        "FailedServer" : "pelican.origin-2.org",
                        "FailureType" : "TooSlow",
                    },
                    {
                        "ErrorType" : "Transfer",
                        "ErrorCode" : 6,
                        "ErrorString" : "Timer expired",
                        "FailedServer" : "pelican.origin-2.org",
                        "FailureType" : "TimedOut",
                    },
                    {
                        "ErrorType" : "Transfer",
                        "ErrorCode" : 7,
                        "ErrorString" : "Exceeded pull rates",
                        "FailedServer" : "pelican.origin-2.org",
                        "FailureType" : "Quota",
                    },
                ],
                "TransferError" : "URL transfer encoded to fail",
            },
            "input",
            13,
            72 << 8,
            None,
            None,
        ),
    ),
    # Check output transfer plugin failure with multiple attempts
    "MULTI_ATTEMPT_FAILURE_OUTPUT" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_output_files" : "foo",
            "transfer_output_remaps" : '"foo=debug://error/contact~[$(contact_output)]/authorization~[$(authorization_output)]/specification~[$(specification_output)]"',
        },
        ClusterState.all_held,
        (
            {
                "TransferProtocol" : "debug",
                "TransferSuccess" : False,
                "TransferFileBytes" : 0,
                "TransferTotalBytes" : 0,
                #"ConnectionTimeSeconds" : 0,
                "TransferErrorData" : [
                    {
                        "ErrorType" : "Contact",
                        "ErrorCode" : 12,
                        "ErrorString" : "Failed to contact server",
                        "FailedServer" : "national.secrets.gov",
                    },
                    {
                        "ErrorType" : "Authorization",
                        "ErrorCode" : 52,
                        "ErrorString" : "Detected bad cookies",
                        "FailedServer" : "cookie.monster.org",
                        "FailureType" : "Authentication",
                        "ShouldRefresh" : True,
                        "DeveloperData" : {
                            "Note" : "Clear cookie cache",
                        },
                    },
                    {
                        "ErrorType" : "Specification",
                        "ErrorCode" : 405,
                        "ErrorString" : "Failed to create file foo",
                        "FailedServer" : "todd.chtc.wisc.edu",
                    },
                ],
                "TransferError" : "URL transfer encoded to fail",
            },
            "output",
            12,
            1 << 8,
            None,
            None,
        ),
    ),
    # Check input transfer plugin success with a failed attempt
    "SUCCESSFUL_INPUT" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_input_files" : 'debug://success/transfer[ErrorCode=1000;ErrorString="Quota%20exceeded";FailedServer="origin-3.osdf.org";FailureType="Quota"]'
        },
        ClusterState.all_complete,
        (
            {
                "TransferProtocol" : "debug",
                "TransferSuccess" : True,
                "TransferFileBytes" : 0,
                "TransferTotalBytes" : 0,
                #"ConnectionTimeSeconds" : 0,
                "TransferErrorData" : [
                    {
                        "ErrorType" : "Transfer",
                        "ErrorCode" : 1000,
                        "ErrorString" : "Quota exceeded",
                        "FailedServer" : "origin-3.osdf.org",
                        "FailureType" : "Quota",
                    },
                ],
            },
            "input",
            0,
            0 << 8,
            { "ExitBySignal": False, "ExitCode": 0 },
            "input",
        ),
    ),
    # Check output transfer plugin success with failed attempt
    "SUCCESSFUL_OUTPUT" : (
        {
            "executable" : "$(path_to_sleep)",
            "arguments" : 0,
            "transfer_output_files" : "foo",
            "transfer_output_remaps" : '"foo=debug://success/specification~[$(specification_output)]"',
        },
        ClusterState.all_complete,
        (
            {
                "TransferProtocol" : "debug",
                "TransferSuccess" : True,
                "TransferFileBytes" : 0,
                "TransferTotalBytes" : 0,
                #"ConnectionTimeSeconds" : 0,
                "TransferErrorData" : [
                    {
                        "ErrorType" : "Specification",
                        "ErrorCode" : 405,
                        "ErrorString" : "Failed to create file foo",
                        "FailedServer" : "todd.chtc.wisc.edu",
                    },
                ],
            },
            "output",
            0,
            0 << 8,
            { "ExitBySignal": False, "ExitCode": 0 },
            "output",
        ),
    ),
}

#--------------------------------------------------------------------------
# List of 'custom submit macro' -> file contents of file encoded ads for debug plugin
PLUGIN_ADS = {
    "parameter_output" :
"""ErrorCode = 32
ErrorString = "Invalid host name provided"
PluginLaunched = True
DeveloperData = [ ConnectionTime=98.3; ]
""",
    "resolution_output" :
"""ErrorCode = 404
ErrorString = "Failed to locate host"
FailedName = "perfect.gaming.host"
FailureType = "Definitive"
DeveloperData = [ ConnectionAttempts = 10; Trolled = True; ]
""",
    "contact_output" :
"""ErrorCode = 12
ErrorString = "Failed to contact server"
FailedServer = "national.secrets.gov"
""",
    "authorization_output" :
"""ErrorCode = 52
ErrorString = "Detected bad cookies"
FailedServer = "cookie.monster.org"
FailureType = "Authentication"
ShouldRefresh = True
DeveloperData = [ Note = "Clear cookie cache"; ]
""",
    "specification_output" :
"""ErrorCode = 405
ErrorString = "Failed to create file foo"
FailedServer = "todd.chtc.wisc.edu"
""",
    "transfer_failure_output" :
"""ErrorCode = 10
ErrorString = "Transfer rate fell below threshold: 100 B/s"
FailedServer = "tj.winblows.net"
FailureType = "TooSlow"
""",
}


@action
def write_plugin_ads(test_dir):
    """Write all debug plugin ad files for encoding. Return dictionary of 'submit macro' -> 'Absolute path filenames'"""
    FILES = dict()

    for TAG, AD in PLUGIN_ADS.items():
        ad_file = test_dir / f"{TAG}.ad"
        ad_file.write_text(AD)
        FILES[TAG] = str(ad_file)

    return FILES

#--------------------------------------------------------------------------
@action
def submit_jobs(default_condor, path_to_sleep, write_plugin_ads):
    """Submit all test cases as individual jobs"""
    HANDLES = dict()

    # Default things to add to all test cases
    defaults = {
        "path_to_sleep" : path_to_sleep,  # Custom macro to reference path to sleep
        "My.TestName" : "$(test_name)",   # Mark test name in job ClassAd
        "log" : "$(test_name).log",       # Set a log file to the test name
        "transfer_executable" : False,    # Don't transfer executable (we have absolute path)
        "should_transfer_files" : True,   # Always transfer files
    }
    # Update defaults with macro -> plugin ad files
    defaults.update(write_plugin_ads)

    # Submit each test job
    for TEST, DETAILS in TEST_CASES.items():
        desc = {}
        desc.update(defaults)
        desc.update({"test_name" : TEST})

        desc.update(DETAILS[0])

        HANDLES[TEST] = default_condor.submit(desc)

    yield HANDLES

#--------------------------------------------------------------------------
@action(params={name: name for name in TEST_CASES})
def test_info(request, submit_jobs):
    return (
        request.param,
        submit_jobs[request.param],
        TEST_CASES[request.param][1],
        TEST_CASES[request.param][2],
    )

@action
def test_name(test_info):
    return test_info[0]

@action
def test_job_handle(test_info):
    return test_info[1]

@action
def test_wait_condition(test_info):
    return test_info[2]

@action
def test_check_details(test_info):
    return test_info[3]

@action
def wait_for_job(test_job_handle, test_wait_condition):
    """Wait for test job based on specified wait condition"""
    assert test_job_handle.wait(condition=test_wait_condition, timeout=300)
    return test_job_handle

#--------------------------------------------------------------------------
@action
def test_get_history(default_condor, test_check_details, wait_for_job):
    """Get the respective test's transfer ad from the epoch history"""
    expected_ad, xfer_type, _, _, _, _ = test_check_details

    # No transfer type specified then skip check
    if xfer_type is None:
        pytest.skip("Skipping check epoch history because transfer type is None")

    # On a busy machine, the log entry can be written before the epoch hit the
    # history file. Retry a few times if we don't see the ad.
    retry_count = 0
    ad = {}
    while (len(ad) == 0 and retry_count < 4):
        p = default_condor.run_command([
            "condor_history",
            "-epochs",
            "-type", xfer_type,
            "-limit", 1,
            "-l",
            wait_for_job.clusterid
            ])
        ad = classad2.parseOne(p.stdout)
        retry_count += 1
        if (len(ad) == 0):
            time.sleep(5)

    return (expected_ad, ad, xfer_type)

#--------------------------------------------------------------------------
@action
def test_get_hold_codes(default_condor, test_check_details, test_wait_condition, wait_for_job):
    """Get the hold code and subcode from test job ad in AP"""
    # If we don't expect the job to be on hold skip this check
    if test_wait_condition != ClusterState.all_held:
        pytest.skip("Skipping check hold codes for non-held job")

    _, _, code, subcode, _, _ = test_check_details

    ads = default_condor.query(
        constraint=f"ClusterId=={wait_for_job.clusterid}",
        projection=["HoldReasonCode", "HoldReasonSubCode"],
        limit=1
    )

    return (code, subcode, ads[0]) if len(ads) > 0 else None

#--------------------------------------------------------------------------
@action
def test_get_plugin_history(default_condor, test_check_details, wait_for_job):
    _, _, _, _, i_ad, i_type = test_check_details

    # No transfer type specified then skip check
    if i_type is None:
        pytest.skip("Skipping check epoch history because transfer type is None")

    # On a busy machine, the log entry can be written before the epoch hit the
    # history file. Retry a few times if we don't see the ad.
    retry_count = 0
    ad = {}
    while (len(ad) == 0 and retry_count < 4):
        p = default_condor.run_command([
            "condor_history",
            "-epochs",
            "-type", i_type,
            "-limit", 1,
            "-l",
            wait_for_job.clusterid
            ])
        ad = classad2.parseOne(p.stdout)
        retry_count += 1
        if (len(ad) == 0):
            time.sleep(5)

    return (i_ad, ad, i_type)

#--------------------------------------------------------------------------
def compare_ads(expected, result):
    """Brute force ad comparison"""
    for key, val in expected.items():
        assert key in result
        if type(val) is dict:
            compare_ads(val, result[key])
        elif type(val) is list:
            for i in range(len(val)):
                compare_ads(val[i], result[key][i])
        else:
            assert val == result[key]

#==========================================================================
class TestPluginFailure:
    def test_check_hold_codes(self, test_get_hold_codes):
        """Check test job hold codes are expected values"""
        assert test_get_hold_codes is not None
        e_code, e_subcode, ad = test_get_hold_codes
        assert e_code == ad.get("HoldReasonCode")
        assert e_subcode == ad.get("HoldReasonSubCode")

    def test_check_epoch_ad(self, test_get_history):
        """Check transfer ad written by shadow has expected values"""
        e_ad, ad, xfer = test_get_history
        xfer_type = f"{xfer.title()}PluginResultList"
        assert xfer_type in ad
        result_ad = ad[xfer_type][0]
        compare_ads(e_ad, result_ad)

    def test_check_plugin_invocation_ad(self, test_get_plugin_history):
        i_ad, ad, i_type = test_get_plugin_history
        invocation_type = f"{i_type.title()}PluginInvocations"
        assert invocation_type in ad
        invocation_ad = ad[invocation_type][0]
        compare_ads(i_ad, invocation_ad)



#!/usr/bin/python3

import sys
from collections import defaultdict

import htcondor2 as htcondor
import classad2 as classad


def main(requested_clusterID):

    constraint = None
    if requested_clusterID is not None:
        constraint = f'clusterID == {requested_clusterID}'
    schedd = htcondor.Schedd()
    results = schedd.jobEpochHistory(
        ad_type=['epoch', 'common'],
        constraint=constraint,
    )


    #
    # Because the shadow can fall back to using normal input transfer for jobs
    # which specify common files, we have to record either (a) that we fell
    # back for a given epoch or (b) that we successfully mapped a catalog for
    # this epoch.  (Otherwise, that is, we can't be sure that if an epoch skipped
    # a transfer because of the common files functionality or not.)
    #
    # So we record (1), each epoch whose job ad defines at least one common
    # file catalog; and (2), each epoch which indicates a succesful mapping.
    #
    # For extra credit, we can correlate COMMON entries to epochs for sizes.
    #


    # For each cluster ID, how many epochs had at least one common file catalog?
    EpochsByClusterID = defaultdict(lambda: 0)
    # For each cluster ID, how many epochs had a common file catalog mapped?
    CommonFilesMappedByClusterID = defaultdict(lambda: 0)
    # For each cluster ID, how many common transfers occurred?
    CommonTransfersByClusterID = defaultdict(lambda: 0)
    # For each cluster ID, how many catalogs were defined?
    CommonCatalogsByClusterID = defaultdict(lambda: 0)
    # For each cluster ID, how many common bytes were transferred?
    CommonBytesByClusterID = defaultdict(lambda: 0)

    for entry in results:
        clusterID = entry['ClusterID']
        epoch_ad_type = entry['EpochAdType']

        if epoch_ad_type == "COMMON":
            CommonTransfersByClusterID[clusterID] += 1

            stats_ad = entry.get('TransferCommonStats')
            if stats_ad is not None:
                protocols = stats_ad['Protocols']
                for protocol in protocols.split(','):
                    bytes = stats_ad[f"{protocol}SizeBytes"]
                    CommonBytesByClusterID[clusterID] += bytes
            else:
                CommonBytesByClusterID[clusterID] += entry['CedarSizeBytes']
        elif epoch_ad_type == "EPOCH":
            old_style = False;
            if 'CommonInputFiles' in entry:
                old_style = True
                CommonCatalogsByClusterID[clusterID] += 1

            new_style = False
            if '_x_common_input_catalogs' in entry:
                new_style = True
                string_list = entry['_x_common_input_catalogs']
                # This isn't quite right, but it's probably close enough.
                list = string_list.split(',')
                CommonCatalogsByClusterID[clusterID] += len(list)

            if old_style or new_style:
                EpochsByClusterID[clusterID] += 1


            if 'CommonFilesMappedTime' in entry:
                CommonFilesMappedByClusterID[clusterID] += 1


    if len(CommonTransfersByClusterID) == 0:
        if requested_clusterID is None:
            print("Found no common file transfers.")
        else:
            print(f"Found no common file transfer for cluster ID {requested_clusterID}.")
        sys.exit(0)


    # We could record the number of succesful mappings, but for now let's just
    # assume that if any mapping in an epoch succeeded, that they all did.
    WholeTransfersByClusterID = defaultdict(lambda: 0)
    BytesPerWholeTransferByClusterID = defaultdict(lambda: 0)
    for clusterID, transfers in CommonTransfersByClusterID.items():
        transfers_per_cluster = CommonCatalogsByClusterID[clusterID]/EpochsByClusterID[clusterID]
        whole_transfers = transfers / transfers_per_cluster
        if float(int(whole_transfers)) != whole_transfers:
            print(
                f"Ignoring partial transfers in cluster ID {clusterID}: "
                f"{transfers} transfers / "
                f"(({CommonCatalogsByClusterID[clusterID]} common catalogs / "
                f"{EpochsByClusterID[clusterID]} epochs) = "
                f"{transfers_per_cluster} transfers_per_cluster) = "
                f"{whole_transfers} whole_transfers"
            )
        whole_transfers = int(whole_transfers)
        WholeTransfersByClusterID[clusterID] = whole_transfers

        bytes_per_whole_transfer = int(
            CommonBytesByClusterID[clusterID] / whole_transfers
        )
        BytesPerWholeTransferByClusterID[clusterID] = bytes_per_whole_transfer

        # And for the money...
        common_epochs = CommonFilesMappedByClusterID[clusterID]
        total_bytes = common_epochs * bytes_per_whole_transfer
        bytes_actually_transferred = whole_transfers * bytes_per_whole_transfer
        pct_a = int(
            ((total_bytes - bytes_actually_transferred)/total_bytes) * 100
        )
        print(
            f"Cluster {clusterID}: required {total_bytes} total bytes in common files "
            f"(as {bytes_per_whole_transfer} bytes per epoch * {common_epochs} epochs, "
            f"not including fall-back epochs), but only transferred "
            f"{bytes_actually_transferred} bytes, skipping "
            f"{total_bytes - bytes_actually_transferred} bytes, or "
            f"{pct_a}% of the total."
        )

    if requested_clusterID is not None:
        return 0


    total_epochs = 0
    for clusterID, count in EpochsByClusterID.items():
        total_epochs += count

    total_common_mappings = 0
    for clusterID, count in CommonFilesMappedByClusterID.items():
        total_common_mappings += count

    total_whole_transfers = 0
    for clusterID, count in WholeTransfersByClusterID.items():
        total_whole_transfers += count


    pct_a = int(
        ((total_common_mappings - total_whole_transfers)/total_common_mappings) * 100
    )
    pct_b = int(
        ((total_epochs - total_whole_transfers)/total_epochs) * 100
    )
    print(
        f"Of {total_epochs} total epochs, "
        f"{total_common_mappings} successfuly mapped common files "
        f"after performing {total_whole_transfers} whole transfers, "
        f"skipping {pct_a}% of those transfers, "
        f"or {pct_b}% of all possible transfers."
    )

    return 0


if __name__ == '__main__':
    clusterID = None
    if len(sys.argv) == 1:
        pass
    elif len(sys.argv) == 2:
        clusterID = int(sys.argv[1])
    else:
        print(f"Usage: {sys.argv[0]} [clusterID]")
        sys.exit(1)
    sys.exit(main(clusterID))



#!/usr/bin/env pytest

import os
import time
import classad2

from pathlib import Path

from ornithology import (
    action,
    write_file,
    format_script,
    Condor,
    ClusterState,
    daemons,
)


import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@action
def the_condor(test_dir):
    local_dir = test_dir / "condor"

    with Condor(
        local_dir = local_dir,
        config = {
            # Strictly speaking, the relevant dprintf() should be
            # duplicated into D_TEST, as well, instead.
            'STARTER_DEBUG':    'D_CATEGORY D_FULLDEBUG',
        },
    ) as the_condor:
        yield the_condor


@action
def the_test_script(test_dir):
    test_script_path = test_dir / "test_script.sh"

    # This shouldn't be hard to conver to Python,
    # which would let us run this test on Windows.
    script = """
        #!/bin/bash
        mkdir -p 1/2/3/4
        touch    1/2/3/4/_file0
        mkdir -p 1/2/a/b
        touch    1/2/a/_file1
        touch    1/2/a/_file2
        touch    1/2/a/b/_file3
        touch    1/2/a/b/_file4
        mkdir -p 1/c/5
        touch    1/c/5/_file5
        mkdir -p 6
        exit 0
    """

    write_file(test_script_path, format_script(script))
    return test_script_path


@action
def the_completed_test_job(the_condor, test_dir, the_test_script):
    description = {
        "executable":               str(the_test_script),
        "should_transfer_files":    "yes",
        "transfer_executable":      "yes",
        "log":                      "log",
        "output":                   "output",
        "error":                    "error",
        "transfer_output_files":    "1",
        "output_destination":       f"null://{test_dir}",
    }

    the_test_job = the_condor.submit(
        description = description,
        count = 1
    )

    the_test_job.wait(
        verbose = True,
        timeout = 60,
        condition = ClusterState.all_complete,
        fail_condition = ClusterState.any_held,
    )

    yield the_test_job

    the_test_job.remove()


class TestPluginsDontSeeDirectories:

    # Theoretically, this could find the wrong log entry, and as such, it
    # it would be better to rewrite the test to use the local plug-in and
    # have it transfer its own input file to someplace we can examine after
    # the test job completes.  But we should really make the local plugin
    # a fixture if we do that, and then change the other tests using it, etc.
    #
    # Also, a bunch of this should be moved into its own fixture because
    # it's actually a set-up failure.
    def test_no_directories_in_input(self,
        the_completed_test_job, the_condor
    ):
        serializedNewClassAds = None
        starter_log_path = the_condor.log_dir / "StarterLog.slot1_1"
        with open(starter_log_path, "r") as log:
            for line in log.readlines():
                if "FILETRANSFER: with plugin input" in line:
                    left_apo = line.find("[")
                    assert left_apo != -1
                    serializedNewClassAds = line[left_apo:-1]
        assert serializedNewClassAds is not None
        print(serializedNewClassAds)

        file_count = 0
        for ad in classad2.parseAds(serializedNewClassAds, classad2.ParserType.New):
            path = ad["LocalFileName"]
            assert not Path(path).is_dir()
            logger.debug(f"Found non-directory '{path}'")
            file_count += 1
        assert(file_count == 6)



import sys
import enum


class ParserType(enum.IntEnum):
    """
    An enumeration of ClassAd parser types.

    .. attribute:: Auto

        Assuming that the input is a sequence of serialized ClassAds
        in either the "new" or the "old" format, determine which.

    .. attribute:: Old

        Assume that the input is a sequence of ClassAds serialized in
        the "old" (or "long") format.

    .. attribute:: New

        Assume that the input is a sequence of ClassAds serialized in
        the "new" format.
    """

    Auto        = -1
    Old         = 0
    XML         = 1
    JSON        = 2
    New         = 3
    FileAuto    = 4



#!/usr/bin/env pytest

import os
import time
from pathlib import Path

from ornithology import (
    action,
    Condor,
    ClusterState,
)


#
# (A) Touch the output files,
# sleep(1) to make sure the timestamps are different,
# then touch the executable,
# then touch the input files.
#
# This should clearly result in the job running.
#
# (B) Touch the output files,
# sleep(1) to make sure the timestamps are different,
# then touch the executable,
# then touch the stdin file,
# then touch the input files.
#
# This should clearly result in the job running.
#


@action
def test_a_dir( test_dir ):
    test_a_dir = test_dir / "test_a"
    test_a_dir.mkdir()

    (test_a_dir / "out1").touch()
    (test_a_dir / "out2").touch()

    time.sleep(1)

    (test_a_dir / "in1").touch()
    (test_a_dir / "in2").touch()

    time.sleep(1)

    (test_a_dir / "exe.sh").touch()
    (test_a_dir / "exe.sh").write_text(
"""#!/bin/bash
touch out1
touch out2
exit 0
"""
    )
    (test_a_dir / "exe.sh").chmod(0o755)

    return test_a_dir


@action
def test_b_dir( test_dir ):
    test_b_dir = test_dir / "test_b"
    test_b_dir.mkdir()

    (test_b_dir / "out1").touch()
    (test_b_dir / "out2").touch()

    time.sleep(1)

    # Avoid triggering the bug from test A by making sure the
    # executable is older than the input files.
    (test_b_dir / "exe.sh").touch()
    (test_b_dir / "exe.sh").write_text(
"""#!/bin/bash
touch out1
touch out2
exit 0
"""
    )
    (test_b_dir / "exe.sh").chmod(0o755)

    time.sleep(1)

    (test_b_dir / "in1").touch()
    (test_b_dir / "in2").touch()

    time.sleep(1)

    (test_b_dir / "stdin.file").touch()

    return test_b_dir


@action
def the_condor(default_condor):
    return default_condor


the_job_description = {
    "LeaveJobInQueue":              "true",

    "executable":                   "exe.sh",
    "transfer_executable":          "true",
    "should_transfer_files":        "true",
    "when_to_transfer_output":      "ON_EXIT",

    "transfer_input_files":         "in1, in2",
    "transfer_output_files":        "out1, out2",

    "skip_if_dataflow":             "true",
}


@action
def completed_test_a_job(test_dir, test_a_dir, the_condor):
    complete_job_description = {
        ** the_job_description,
        ** {
            "log": (test_dir / "test_a.log").as_posix(),
        }
    }

    os.chdir(test_a_dir)
    the_handle = the_condor.submit(
        complete_job_description,
        count=1,
    )

    assert the_handle.wait(
        timeout=60,
        condition=ClusterState.all_complete,
        fail_condition=ClusterState.any_held
    )

    return the_handle


@action
def completed_test_b_job(test_dir, test_b_dir, the_condor):
    complete_job_description = {
        ** the_job_description,
        ** {
            "log": (test_b_dir / "test_b.log").as_posix(),
            "input": (test_b_dir / "stdin.file").as_posix(),
        }
    }

    os.chdir(test_b_dir)
    the_handle = the_condor.submit(
        complete_job_description,
        count=1,
    )

    assert the_handle.wait(
        timeout=60,
        condition=ClusterState.all_complete,
        fail_condition=ClusterState.any_held
    )

    return the_handle


class TestDataflowMore:

    def test_a_skipped_job(self, the_condor, completed_test_a_job):
        schedd = the_condor.get_local_schedd()
        constraint = f'ClusterID == {completed_test_a_job.clusterid}'
        result = schedd.query(
            constraint=constraint,
            projection=["DataflowJobSkipped"],
        )
        skipped = result[0]["DataflowJobSkipped"]
        assert(not skipped)


    def test_b_skipped_job(self, the_condor, completed_test_b_job):
        schedd = the_condor.get_local_schedd()
        constraint = f'ClusterID == {completed_test_b_job.clusterid}'
        result = schedd.query(
            constraint=constraint,
            projection=["DataflowJobSkipped"],
        )
        skipped = result[0]["DataflowJobSkipped"]
        assert(not skipped)



import os
import html
import sys

from docutils import nodes, utils
from docutils.parsers.rst import Directive
from sphinx import addnodes
from sphinx.errors import SphinxError
from sphinx.util.nodes import split_explicit_title, process_index_entry, set_role_source_info
from htc_helpers import *

CATEGORIES = []

def dump(obj):
    for attr in dir(obj):
        print("obj.%s = %r" % (attr, getattr(obj, attr)))

def conf_temp_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    global CATEGORIES
    # Create a new linkable target using the config template
    item, info = custom_ext_parser(text, "<", ">")
    info = info.upper()

    # The category sections don't expect any extra information i.e. POLICY, FEATURE, etc
    if info == "":
        targetid = item.upper()
        index_cat = "Configuration Templates"
        if item.upper() in CATEGORIES:
            warn(f"Config template category '{item}' defined multiple times.")
        else:
            CATEGORIES.append(item.upper())
        node = nodes.target('', item, ids=[targetid], classes=['config-template'])
    # Handle specific template (FEATURE:GPUs)
    else:
        if info not in CATEGORIES:
            warn(f"Provided category '{info}' for config template '{item}' is not defined.")
        # Find optional function parts| FOO:Bar(ignore this stuff)
        display_text = item
        item = item[:item.find("(")] if "(" in item else item
        targetid = info + ":" + item.upper()
        index_cat = f"{info} Configuration Templates"
        html_parser = html.parser.HTMLParser()
        html_markup = f"""<code class="docutils literal notranslate"><span id="{targetid}" class="pre">{html.escape(display_text)}</span></code>"""
        node = nodes.raw("", html_markup, format="html")

    # Automatically include an index entry for config templates
    indexnode = addnodes.index()
    indexnode['entries'] = process_index_entry(f"pair: {item}; {index_cat}", targetid)
    set_role_source_info(inliner, lineno, indexnode)

    headerlink_node = make_headerlink_node(targetid, options)

    nodes_list = [headerlink_node, indexnode, node] if info == "" else [indexnode, node, headerlink_node]

    return nodes_list, []


def setup(app):
    app.add_role("config-template", conf_temp_role)




import htcondor2
from htcondor2._utils.ansi import underline

from htcondor_cli.convert_ad import ads_to_daemon_status, health_str
from htcondor_cli.noun import Noun
from htcondor_cli.verb import Verb


class Status(Verb):
    """
    Displays status information all central managers defined by COLLECTOR_HOST
    """

    options = {
        # Verb specific CLI options/flags
    }

    def __init__(self, logger, **options):
        constraint = '(MyType == "Negotiator" || MyType == "Collector")'
        ads = []

        collectors = htcondor2.param["COLLECTOR_HOST"].replace(" ", ",").replace("\t", ",")
        for host in collectors.split(","):
            if host == "":
                continue

            try:
                collector = htcondor2.Collector(host)
                ads.extend(collector.query(constraint=constraint))
            except:
                logger.info(f"Failed to query Collector at {host}")

        if len(ads) == 0:
            raise RuntimeError("Failed to locate any Central Managers")

        cm_status = {}
        width = 0

        for daemon, status in ads_to_daemon_status(ads):
            host = status["Machine"]
            width = len(host) if len(host) > width else width

            if host in cm_status:
                cm_status[host]["HP"] += status["HealthPoints"]
                cm_status[host]["Count"] += 1
            else:
                cm_status[host] = {"HP": status["HealthPoints"], "Count": 1}

        columns = ["Central Manager", "Health"]
        logger.info(underline("{0: ^{width}}   {1}".format(*columns, width=width)))
        for host, status in cm_status.items():
            logger.info(f"{host: <{width}}   {health_str(status['HP'], status['Count'])}")
        logger.info("")


class CentralManager(Noun):
    """
    Run operations on pool central managers
    """

    class status(Status):
        pass

    @classmethod
    def verbs(cls):
        return [cls.status]



import os
import sys

from docutils import nodes
from docutils.parsers.rst import Directive
from sphinx import addnodes
from sphinx.errors import SphinxError
from sphinx.util.nodes import split_explicit_title, process_index_entry, set_role_source_info
from htc_helpers import *

SUBMIT_CMDS = []

def find_submit_cmds(dir: str):
    subcoms = []
    submit_man = os.path.join(dir, "man-pages", "condor_submit.rst")
    with open(submit_man, "r", encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            for subcom in get_all_defined_role("subcom-def", line):
                if subcom not in subcoms:
                    subcoms.append(subcom)
    subcoms.sort()
    return subcoms

def dump(obj):
    for attr in dir(obj):
        print("obj.%s = %r" % (attr, getattr(obj, attr)))

def subcom_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    root_dir = root_dir = get_rel_path_to_root_dir(inliner)[:-1]
    subcom_name, subcom_index = custom_ext_parser(text)
    if subcom_name not in SUBMIT_CMDS:
        docname = inliner.document.settings.env.docname
        warn(f"{docname}:{lineno} | Submit command '{subcom_name}' not found in defined list. Either a typo or not defined.")
    ref_link = f"href=\"{root_dir}/man-pages/condor_submit.html#" + str(subcom_name) + "\""
    return make_ref_and_index_nodes(name, subcom_name, subcom_index,
                                    ref_link, rawtext, inliner, lineno, options)

def setup(app):
    global SUBMIT_CMDS
    SUBMIT_CMDS = find_submit_cmds(app.srcdir)
    app.add_role("subcom", subcom_role)




#!/usr/bin/env pytest

import logging
import os
from time import sleep
from pathlib import Path
import htcondor2 as htcondor

from ornithology import (
    config,
    standup,
    action,
    Condor,
    ClusterState,
    ClusterHandle,
    JobStatus
)


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

@standup
def condor(test_dir):
    with Condor(
        local_dir=test_dir / "condor",
        raw_config="""
            SCHEDD_INTERVAL = 5
            PERIODIC_EXPR_INTERVAL = 5
        """,
    ) as condor:
        yield condor

@action
def remote_job(condor, path_to_sleep):
    schedd = condor.get_local_schedd()
    jdl = htcondor.Submit(
        f"""
        executable = {path_to_sleep}
        arguments = 0
        log = the_job.log
        """
        )
    sr = schedd.submit(jdl, spool=True)
    schedd.spool(sr)

    # This is a hack, although it duplicates what the_condor.submit() would
    # do if that function supported spooling.
    ch = ClusterHandle(condor, sr)
    # This is a really egregious hack because the EventLog class (and company)
    # don't understand that a log file with a relative path may mean "in SPOOL".
    ch.event_log._event_log_path = Path(condor.spool_dir / f"{sr.cluster()}" / "0" / f"cluster{sr.cluster()}.proc0.subproc0" / "the_job.log")

    assert ch.wait(condition=ClusterState.all_complete, timeout=60)
    schedd.retrieve(sr.cluster())

    return ch

class TestRemoteSubmit:
    def test_remote_submit_retirement(self, condor, remote_job):
        schedd = condor.get_local_schedd()
        # Poke the schedd to retire the job more quickly
        schedd.reschedule()

        in_queue = True
        i = 0
        while i < 60 and in_queue:
            ads = schedd.query()
            in_queue = len(ads) > 0
            i += 1
            sleep(1)

        assert not in_queue



#!/usr/bin/env pytest

import pytest

import os
from pathlib import Path
import subprocess

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

from ornithology import (
    action,
)


TEST_CASES = {
    "on_and_on": {
        "env": {
            '_CONDOR_CONTAINER_IMAGES_COMMON_BY_DEFAULT': 'TRUE'
        },
        "sub": "container_is_common = true",
        "expected": "_x_catalog_condor_container_image",
    },
    "on_and_off": {
        "env": {
            '_CONDOR_CONTAINER_IMAGES_COMMON_BY_DEFAULT': 'TRUE'
        },
        "sub": "container_is_common = false",
        "expected": "TransferInput",
    },
    "on_and_unspec": {
        "env": {
            '_CONDOR_CONTAINER_IMAGES_COMMON_BY_DEFAULT': 'TRUE'
        },
        "sub": "",
        "expected": "_x_catalog_condor_container_image",
    },

    "off_and_on": {
        "env": {
            '_CONDOR_CONTAINER_IMAGES_COMMON_BY_DEFAULT': 'TRUE'
        },
        "sub": "container_is_common = true",
        "expected": "_x_catalog_condor_container_image",
    },
    "off_and_off": {
        "env": {
            '_CONDOR_CONTAINER_IMAGES_COMMON_BY_DEFAULT': 'FALSE'
        },
        "sub": "container_is_common = false",
        "expected": "TransferInput",
    },
    "off_and_unspec": {
        "env": {
            '_CONDOR_CONTAINER_IMAGES_COMMON_BY_DEFAULT': 'FALSE'
        },
        "sub": "",
        "expected": "TransferInput",
    },
}


@action(params={name: name for name in TEST_CASES})
def the_test_pair(request):
    return request.param, TEST_CASES[request.param]


@action
def the_test_name(the_test_pair):
    return the_test_pair[0]


@action
def the_test_case(the_test_pair):
    return the_test_pair[1]


@action
def the_test_result(the_test_name, the_test_case):
    subtext = f"""
        universe = vanilla
        container_image = busybox.sif
        shell = ls -la
        should_transfer_files = YES
        {the_test_case['sub']}
        queue
    """

    env = {
        ** os.environ,
        ** the_test_case['env'],
    }

    rv = subprocess.run(
        ['condor_submit', '-', '-dry', '-'],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        input=subtext,
        universal_newlines=True,
        env=env,
        timeout=2,
    )

    for line in rv.stdout.splitlines():
        if 'ContainerImage' in line:
            continue
        if 'busybox.sif' in line:
            return line.split('=')[0]


class TestCIFKnob:

    def test_correct_job_ad(self, the_test_case, the_test_result):
        assert the_test_result == the_test_case['expected']



#!/usr/bin/env pytest

# test_dagman_job_list_failure_tolerance.py
#
# This test checks that the DAGMan job list fault tolerance
# works correctly for nodes

from ornithology import *
import htcondor2
import os
import json
import enum
from pathlib import Path
from shutil import rmtree
from time import time as now

TIMEOUT = 120

DAG_FILENAME = "test.dag"
DAG_CONTENTS = """
SUBMIT-DESCRIPTION chain @=desc
    executable = {exe}
    arguments  = $(ClusterId) $(ProcId) {tolerance}
    # Possible race condition with local universe may cause
    # abort event for jobs that have terminated upon DAGMan exit
    #universe   = local
    priority   = $(ProcId) * -100

    getenv     = True

    queue {count}
@desc

JOB TEST chain
"""

#------------------------------------------------------------------
# DAG Status values (keep in sync with enum in dagman_utils.h)
class DagStatus(enum.IntEnum):
    OK = 0
    ERROR = 1
    NODE_FAILED = 2
    ABORT = 3
    #RM = 4
    #CYCLE = 5
    #HALTED = 6

# Class to hold each test case:
#   - queue: # jobs to queue in list
#   - tolerance: Job failure tolerance before failure
#   - code: Expected DAGMan job exit code
#   - status: Expected DAG Status value
class TestCase:
    __test__ = False
    def __init__(self, queue: int, tolerance: int, code: int, status: DagStatus):
        self.queue = queue
        self.tolerance = tolerance
        self.exit_code = code
        self.status = status

#==================================================================
TEST_CASES = {
    "DEFAULT": TestCase(3, 0, 1, DagStatus.NODE_FAILED),
    "ONE_FAILURE": TestCase(3, 1, 1, DagStatus.NODE_FAILED),
    "FIVE_FAILURES": TestCase(10, 5, 1, DagStatus.NODE_FAILED),
    "NINE_FAILURES": TestCase(10, 9, 1, DagStatus.NODE_FAILED),
    "TOLERATE_ALL": TestCase(3, 3, 0, DagStatus.OK),
    "GREATER_THAN_NUM_PROCS": TestCase(2, 100, 0, DagStatus.OK),
}

#------------------------------------------------------------------
@action
def write_exe(test_dir):
    executable = test_dir / "chain.py"
    with open(executable, "w") as f:
        f.write(
"""#!/usr/bin/env python3

import htcondor2
import sys

FAILURE_CODE = 255
EXIT_CODE = 1

NUM_ARGS = len(sys.argv)
if NUM_ARGS != 4:
    print(f"ERROR: Invlaid number of arguments: {NUM_ARGS}")
    sys.exit(FAILURE_CODE)

try:
    CLUSTER = int(sys.argv[1])
except Exception as e:
    print(f"ERROR: Invalid ClusterId provided: {e}")
    sys.exit(FAILURE_CODE)

try:
    PROC = int(sys.argv[2])
except Exception as e:
    print(f"ERROR: Invalid ProcId provided: {e}")
    sys.exit(FAILURE_CODE)

try:
    THRES = int(sys.argv[3])
except:
    print(f"ERROR: Invalid ProcId provided: {e}")
    sys.exit(FAILURE_CODE)


if PROC == 0:
    sys.exit(EXIT_CODE)

ads = htcondor2.Schedd().query(
    constraint=f'ClusterId=={CLUSTER}&&ProcId=={PROC}',
    projection=["DAGManNodesLog"]
)

if len(ads) != 1:
    print(f"ERROR: Failed to get log(s) for job {CLUSTER}.{PROC}")
    sys.exit(FAILURE_CODE)

LOG = ads[0].get("DAGManNodesLog")

if LOG is None:
    print(f"ERROR: Job {CLUSTER}.{PROC} has no logs defined")
    sys.exit(FAILURE_CODE)

JEL = htcondor2.JobEventLog(LOG)

DONE_EVENTS = [
    htcondor2.JobEventType.JOB_TERMINATED,
    htcondor2.JobEventType.JOB_ABORTED,
]

while True:
    if PROC <= THRES:
        for event in JEL.events(stop_after=1):
            if event.cluster == CLUSTER and event.proc == (PROC - 1) and event.type in DONE_EVENTS:
                sys.exit(EXIT_CODE)
""")

    os.chmod(str(executable), 0o0777)
    return executable

#------------------------------------------------------------------
@action
def run_dags(default_condor, test_dir, write_exe):
    TESTS = dict()

    for TEST, INFO in TEST_CASES.items():
        DAG_MODIFIERS = {
            "exe" : write_exe,
            "tolerance": INFO.tolerance,
            "count": INFO.queue,
        }

        os.chdir(str(test_dir))
        case_dir = test_dir / TEST
        if case_dir.exists():
            rmtree(case_dir)
        os.mkdir(str(case_dir))
        os.chdir(str(case_dir))

        with open(DAG_FILENAME, "w") as f:
            f.write(DAG_CONTENTS.format(**DAG_MODIFIERS))

        DAG = htcondor2.Submit.from_dag(DAG_FILENAME, {"AddToEnv": f"_CONDOR_DAGMAN_NODE_JOB_FAILURE_TOLERANCE={INFO.tolerance}"})
        HANDLE = default_condor.submit(DAG)
        TESTS[TEST] = (HANDLE, INFO.tolerance, INFO.exit_code, INFO.status, INFO.queue)

    yield TESTS

#------------------------------------------------------------------
@action(params={name: name for name in TEST_CASES})
def test_info(request, run_dags, test_dir):
    return (
        request.param,               # Test name
        run_dags[request.param][0],  # Test DAG job handle
        run_dags[request.param][1],  # Test failure tolerance
        run_dags[request.param][2],  # Test expected exit code
        run_dags[request.param][3],  # Test expected DAG status
        run_dags[request.param][4],  # Test queued # of jobs
    )

@action
def test_name(test_info):
    return test_info[0]

@action
def test_handle(test_info):
    return test_info[1]

@action
def test_tolerance(test_info):
    return test_info[2]

@action
def test_exit_code(test_info):
    return test_info[3]

@action
def test_dag_status(test_info):
    return test_info[4]

@action
def test_num_procs(test_info):
    return test_info[5]

#==================================================================
# Magical class to cd into test specific dir and return to root test dir
# via pythons 'with' keyword
class ChangeDir:
    def __init__(self, switch):
        self.orig = None
        self.switch = switch
    def __enter__(self):
        self.orig = os.getcwd()
        os.chdir(self.switch)
    def __exit__(self, exec_type, exec_value, exec_traceback):
        os.chdir(self.orig)
        self.orig = None

#------------------------------------------------------------------
class TestDAGManJobFailTolerance:
    def test_dag_wait(self, test_handle):
        assert test_handle.wait(condition=ClusterState.all_complete, timeout=TIMEOUT)

    def test_dag_final_status(self, test_name, test_exit_code, test_dag_status):
        with ChangeDir(test_name):
            with open(f"{DAG_FILENAME}.metrics", "r") as f:
                metrics = json.load(f)
                assert metrics["exitcode"] == test_exit_code
                assert metrics["DagStatus"] == int(test_dag_status)

    def test_job_proc_states(self, test_name, test_tolerance, test_num_procs):
        with ChangeDir(test_name):
            LOG = Path(f"{DAG_FILENAME}.nodes.log")
            assert LOG.exists()

            START = now()
            PROCS = {
                "submitted": 0,
                "completed": 0,
            }

            with htcondor2.JobEventLog(str(LOG)) as JEL:
                while True:
                    assert now() - START < TIMEOUT
                    for event in JEL.events(stop_after=1):
                        if event.type == htcondor2.JobEventType.SUBMIT:
                            PROCS["submitted"] += 1
                            assert PROCS["submitted"] <= test_num_procs
                        elif event.type == htcondor2.JobEventType.JOB_TERMINATED:
                            assert event.proc <= test_tolerance
                            PROCS["completed"] += 1
                        elif event.type == htcondor2.JobEventType.JOB_ABORTED:
                            assert event.proc > test_tolerance
                            PROCS["completed"] += 1

                    if PROCS["completed"] == test_num_procs:
                        break




#!/usr/bin/env pytest

import htcondor2

import logging
from getpass import getuser

from ornithology import *


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


#
# This test is woefully incomplete: it _assumes_ that the various credential
# verbs have their intended effect, or are effected as intended, on (or by)
# the rest of the system.  Elimnating that assumption would be a lot of work,
# but would also make it possible to write verb-specific tests -- although
# the complete set of tests would still need to check that the command-line
# tool works properly with itself.
#


@action
def local_dir(test_dir):
    return test_dir / "condor"


@action
def the_condor(local_dir):
    with Condor(
        local_dir=local_dir,
        raw_config="""
            # Enable OAuth2.
            SEC_CREDENTIAL_DIRECTORY_OAUTH      = $(LOCAL_DIR)/oauth.d
            CREDMON_OAUTH_LOG                   = $(LOG)/CredMonOAuthLog
            DAEMON_LIST                         = $(DAEMON_LIST),CREDMON_OAUTH
            AUTO_INCLUDE_CREDD_IN_DAEMON_LIST   = True
            TRUST_CREDENTIAL_DIRECTORY          = True
            LOCAL_CREDMON_PROVIDER_NAME         = scitokens
            LOCAL_CREDMON_TOKEN_AUDIENCE        = https://localhost
            ALLOW_DAEMON                        = *
            LOCAL_CREDMON_PRIVATE_KEY           = $(LOCAL_DIR)/trust_domain_ca_privkey.pem
            LOCAL_CREDMON_PUBLIC_KEY            = $(LOCAL_DIR)/trust_domain_ca_pubkey.pem
        """,
    ) as the_condor:
        yield the_condor


@action
def fake_refresh_token_file(test_dir):
    token_file = test_dir / "the_fake_refresh_token"
    token_file.write_text("fake-refresh-token")
    return token_file


@action
def plain_token_file(test_dir):
    token_file = test_dir / "the_plain_token"
    token_file.write_text("plain-token")
    return token_file


class TestHTCondorCredential:

    def test_local_issuer_credential(self, local_dir, the_condor, fake_refresh_token_file):
        rv = the_condor.run_command(
            ['htcondor', '-v', 'credential', 'add', 'oauth2',
             str(fake_refresh_token_file),
             '--service', 'scitokens'],
        )
        assert rv.returncode == 0
        logger.info(rv.stdout)
        logger.info(rv.stderr)

        daemon_log = the_condor.credmon_oauth_log.open()
        assert daemon_log.wait(
            condition=lambda line: "Successfully renewed" in line.message,
            timeout=20,
        )

        rv = the_condor.run_command(
            ['htcondor', '-v', 'credential', 'get', 'oauth2',
            '--service', 'scitokens'],
        )
        assert rv.returncode == 0
        logger.info(rv.stdout)
        logger.info(rv.stderr)

        use_file = local_dir / "oauth.d" / getuser() / "scitokens.use"
        assert rv.stdout == use_file.read_text()

    def test_plain_credential(self, local_dir, the_condor, plain_token_file):
        rv = the_condor.run_command(
            ['htcondor', '-v', 'credential', 'add', 'oauth2',
             str(plain_token_file),
             '--service', 'plain'],
        )
        assert rv.returncode == 0
        logger.info(rv.stdout)
        logger.info(rv.stderr)

        rv = the_condor.run_command(
            ['htcondor', '-v', 'credential', 'get', 'oauth2',
            '--service', 'plain'],
        )
        assert rv.returncode == 0
        logger.info(rv.stdout)
        logger.info(rv.stderr)

        use_file = local_dir / "oauth.d" / getuser() / "plain.use"
        assert rv.stdout == use_file.read_text()



#!/usr/bin/env pytest

#testreq: personal
"""<<CONDOR_TESTREQ_CONFIG
	# make sure that file transfer plugins are enabled (might be disabled by default)
	ENABLE_URL_TRANSFERS = true
"""
#endtestreq

import logging
import os

from ornithology import *
from pathlib import Path

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@action
def tor_job(default_condor, path_to_sleep, test_dir):
    return default_condor.submit(
        {
            "executable":               path_to_sleep,
            "transfer_executable":      "False",
            "should_transfer_files":    "True",

            "log":                      test_dir / "tor_job.log",
            "hold":                     "True",

            "output":                   "output",
            "transfer_output_remaps":   '"output=badname://foo/bar"',
        },
        count=1,
    )


@action
def tor_job_requirements(tor_job):
    return str(tor_job.clusterad["Requirements"])


@action
def nph_job(default_condor, path_to_sleep, test_dir):
    # The idea here is that the job will fail the first time because the
    # output remap points to a bad URL. This used to cause the job to be held, but
    # with the fix in HTCONDOR-2952 the job should just go back to idle in the event 
    # that an EP failed to launch a transfer plugin. The second time the job runs, it should 
    # succeed because the remap expression is designed to only point to the bad URL on the first attempt.
    handle = default_condor.submit(
        {
            "executable":               path_to_sleep,
            "transfer_executable":      "False",
            "should_transfer_files":    "True",
            "arguments":                1,

            "log":                      test_dir / "nph_job.log",

            "transfer_output_files":    path_to_sleep,
            "+TransferOutputRemaps":   f'NumJobStarts > 0 ? "" :  "{path_to_sleep}=badname://foo/bar"',
            "requirements":             'TARGET.HasFileTransferPluginMethods =!= undefined',
        },
        count=1,
    )
    return handle, default_condor.get_local_schedd()

@action
def nph_input_job(default_condor, path_to_sleep, test_dir):
    # The idea here is that the job will fail the first time because the
    # input points to a bad URL. This used to cause the job to be held, but
    # with the fix in HTCONDOR-2952 the job should just go back to idle in the event 
    # that an EP failed to launch a transfer plugin. The second time the job runs, it should 
    # succeed because the TransferInput expression is designed to only point to the bad URL on the first attempt.
    handle = default_condor.submit(
        {
            "executable":               path_to_sleep,
            "transfer_executable":      "False",
            "should_transfer_files":    "True",
            "arguments":                1,

            "log":                      test_dir / "nph_input_job.log",

            "+TransferInput":   f'NumShadowStarts > 1 ? "" :  "badname://foo/bar"',
            "requirements":             'TARGET.HasFileTransferPluginMethods =!= undefined',
        },
        count=1,
    )
    return handle, default_condor.get_local_schedd()

@action
def missing_result_input(default_condor, path_to_sleep, test_dir):
    # The idea here is that the job will fail the first time because the
    # plugin will not return an result file. This used to cause the job to be held, but
    # with the fix in HTCONDOR-2952 the job should just go back to idle in the event 
    # that an EP failed to launch a transfer plugin. The second time the job runs, it should 
    # succeed because the TransferInput expression is designed to only point to the bad URL on the first attempt.
    handle = default_condor.submit(
        {
            "executable":               path_to_sleep,
            "transfer_executable":      "False",
            "should_transfer_files":    "True",
            "arguments":                1,

            "log":                      test_dir / "missing_result_input.log",

            "+TransferInput":   f'NumShadowStarts > 1 ? "" :  "debug://exitnoad/0"',
        },
        count=1,
    )
    return handle, default_condor.get_local_schedd()

'''
This test is commented out for now as it results in the shadow/starter disconnecting multiple times... will fix at
a later date.
@action
def missing_result_output(default_condor, path_to_sleep, test_dir):
    # The idea here is that the job will fail the first time because the
    # plugin will not return an result file. This used to cause the job to be held, but
    # with the fix in HTCONDOR-2952 the job should just go back to idle in the event 
    # that an EP failed to launch a transfer plugin. The second time the job runs, it should 
    # succeed because the TransferInput expression is designed to only point to the bad URL on the first attempt.
    handle = default_condor.submit(
        {
            "executable":               path_to_sleep,
            "transfer_executable":      "False",
            "should_transfer_files":    "True",
            "arguments":                1,

            "log":                      test_dir / "missing_result_output.log",

            "transfer_output_files":    path_to_sleep,
            "+TransferOutputRemaps":   f'NumJobStarts > 0 ? "" :  "{path_to_sleep}=debug://exitnoad/0"',
        },
        count=1,
    )
    return handle, default_condor.get_local_schedd()
'''

@action
def bad_result_input(default_condor, path_to_sleep, test_dir):
    # The idea here is that the job will fail the first time because the
    # plugin will not return a valid result file. This used to cause the job to be held, but
    # with the fix in HTCONDOR-2952 the job should just go back to idle in the event 
    # that an EP failed to launch a transfer plugin. The second time the job runs, it should 
    # succeed because the TransferInput expression is designed to only point to the bad URL on the first attempt.
    handle = default_condor.submit(
        {
            "executable":               path_to_sleep,
            "transfer_executable":      "False",
            "should_transfer_files":    "True",
            "arguments":                1,

            "log":                      test_dir / "bad_result_input.log",

            "+TransferInput":   f'NumShadowStarts > 1 ? "" :  "debug://exitbadad/0"',
        },
        count=1,
    )
    return handle, default_condor.get_local_schedd()

@action
def trailing_semi_job(default_condor, test_dir):
    os.mkdir("sandbox")
    handle = default_condor.submit(
        {
            "executable":               "/usr/bin/touch",
            "transfer_executable":      "False",
            "should_transfer_files":    "True",
            "arguments":                "one two three",

            "log":                      test_dir / "trailing_semi_job.log",

            "transfer_output_remaps":   '"one=sandbox/job_one;two=sandbox/job_two;three=sandbox/job_three;"',
            "output":                   "job_stdout",
            "error":                    "job_stderr",
        },
        count=1,
    )
    return handle


class TestFTPluginStuff:

    def test_tor_requirements(self, tor_job_requirements):
        # How specific do we have to get here?
        assert 'badname' in tor_job_requirements

    def test_no_remap_plugin_vacate(self, nph_job):
        nph_job[0].wait(
            condition=ClusterState.all_complete,
            fail_condition=ClusterState.any_held,
            timeout=60,
        )
        # Verify the job did not go on hold, retried output transfer after plugin failure, and completed successfully.
        assert nph_job[0].state.all_complete()
        # Verify the job went through a vacate cycle with the expected subcode for a missing plugin (-1001)
        assert nph_job[1].jobEpochHistory(
            constraint=f'ClusterID == {nph_job[0].clusterid} && ProcID == 0',
            projection=["VacateReasonSubCode"],
        )[1]["VacateReasonSubCode"] == -1001

    def test_no_input_plugin_vacate(self, nph_input_job):
        nph_input_job[0].wait(
            condition=ClusterState.all_complete,
            fail_condition=ClusterState.any_held,
            timeout=60,
        )
        # Verify the job did not go on hold, retried input transfer after plugin failure, and completed successfully.
        assert nph_input_job[0].state.all_complete()
        # Verify the job went through a vacate cycle with the expected subcode for a missing plugin (-1001)
        assert nph_input_job[1].jobEpochHistory(
            constraint=f'ClusterID == {nph_input_job[0].clusterid} && ProcID == 0',
            projection=["VacateReasonSubCode"],
        )[1]["VacateReasonSubCode"] == -1001

    def test_input_no_plugin_results_vacate(self, missing_result_input):
        missing_result_input[0].wait(
            condition=ClusterState.all_complete,
            fail_condition=ClusterState.any_held,
            timeout=60,
        )
        # Verify the job did not go on hold, retried input transfer after plugin failure, and completed successfully.
        assert missing_result_input[0].state.all_complete()
        # Verify the job went through a vacate cycle with the expected subcode for a missing plugin results file (-1004)
        assert missing_result_input[1].jobEpochHistory(
            constraint=f'ClusterID == {missing_result_input[0].clusterid} && ProcID == 0',
            projection=["VacateReasonSubCode"],
        )[1]["VacateReasonSubCode"] == -1004

    '''
    This test is commented out for now as it results in the shadow/starter disconnecting multiple times... will fix at
    a later date.
    def test_output_no_plugin_results_vacate(self, missing_result_output):
        missing_result_output[0].wait(
            condition=ClusterState.all_complete,
            fail_condition=ClusterState.any_held,
            timeout=60,
        )
        # Verify the job did not go on hold, retried output transfer after plugin failure, and completed successfully.
        assert missing_result_output[0].state.all_complete()
        # Verify the job went through a vacate cycle with the expected subcode for a missing plugin results file (-1004)
        assert missing_result_output[1].jobEpochHistory(
            constraint=f'ClusterID == {missing_result_output[0].clusterid} && ProcID == 0',
            projection=["VacateReasonSubCode"],
        )[1]["VacateReasonSubCode"] == -1004
    '''

    def test_input_bad_plugin_results_vacate(self, bad_result_input):
        bad_result_input[0].wait(
            condition=ClusterState.all_complete,
            fail_condition=ClusterState.any_held,
            timeout=60,
        )
        # Verify the job did not go on hold, retried input transfer after plugin failure, and completed successfully.
        assert bad_result_input[0].state.all_complete()
        # Verify the job went through a vacate cycle with the expected subcode for a missing plugin results file (-1004)
        assert bad_result_input[1].jobEpochHistory(
            constraint=f'ClusterID == {bad_result_input[0].clusterid} && ProcID == 0',
            projection=["VacateReasonSubCode"],
        )[1]["VacateReasonSubCode"] == -1004

    def test_trailing_semi_remap(self, trailing_semi_job):
        trailing_semi_job.wait(
            condition=ClusterState.any_complete,
            fail_condition=ClusterState.all_held,
            timeout=60,
        )
        assert os.path.isfile("job_stdout")
        assert os.path.isfile("job_stderr")
        assert os.path.isfile("sandbox/job_one")
        assert os.path.isfile("sandbox/job_two")
        assert os.path.isfile("sandbox/job_three")



#!/usr/bin/env pytest

import pytest

import os
import tempfile
import subprocess

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def test_filter_exe_results():
    rv = subprocess.run(["test_filter.exe"],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        timeout=20)
    assert rv.returncode == 0



#!/usr/bin/env pytest

#testreq: personal
"""<<CONDOR_TESTREQ_CONFIG
	DAGMAN_DEBUG = D_TEST D_CAT
"""
#endtestreq

from ornithology import *
import htcondor2
import pytest
import os
import sys
import json

from time import sleep
from pathlib import Path
from shutil import rmtree
from typing import Callable, Tuple, Any

#-----------------------------------------------------------------------------------------
PYTHON3 = sys.executable
TIMEOUT = 120
DAG_FILENAME = "test.dag"

#-----------------------------------------------------------------------------------------
def script_check_retry(path):
    script = path / "check_retry.py"
    with open(str(script), "w") as f:
        f.write("""
import sys

EXPECTED = int(sys.argv[1])
RETRY_MAX = int(sys.argv[2])

if EXPECTED != RETRY_MAX:
    sys.exit(125)

sys.exit(0)
""")

    return script


def script_simple_exit(path):
    script = path / "exit.py"
    with open(str(script), "w") as f:
        f.write("""
import sys

CODE = 125
try:
    if len(sys.argv) >= 2:
        CODE = int(sys.argv[1])
except Exception as e:
    print(f"ERROR: Failed to convert specified exit code: {e}")

sys.exit(CODE)
""")

    return script


def write_dag_file(contents: str, name: str = DAG_FILENAME) -> str:
    with open(name, "w") as f:
        f.write(contents)
    return name

#-----------------------------------------------------------------------------------------
def setup_special_nodes(path: Path) -> str:
    exe = script_simple_exit(path)
    return write_dag_file(f"""
SUBMIT-DESCRIPTION test @=desc
    executable = {PYTHON3}
    arguments  = {exe} $(code)
@desc

JOB A test
SERVICE S test
FINAL F test

VARS S code=0
VARS F code=0

# This should not overwrite special node types
VARS ALL_NODES code=1

# Ensure worker node does not fail
VARS A code=0

ABORT-DAG-ON S 1
""")

#-----------------------------------------------------------------------------------------
def setup_all_nodes_vars(path: Path) -> str:
    exe = script_simple_exit(path)
    return write_dag_file(f"""
SUBMIT-DESCRIPTION test @=desc
    executable = {PYTHON3}
    arguments  = {exe} $(code)
    transfer_executable = false
    queue 1
@desc

JOB A test
JOB B test
JOB C test

VARS A code=1
VARS B code=1
VARS C code=1

# Override VARS code variable so jobs execute successfully
VARS ALL_NODES code=0
""")

#-----------------------------------------------------------------------------------------
def setup_all_nodes_scripts(path: Path) -> str:
    exe = script_simple_exit(path)
    return write_dag_file(f"""

SUBMIT-DESCRIPTION test @=desc
    executable = {PYTHON3}
    arguments  = {exe} 0
    transfer_executable = false
    queue 1
@desc

JOB A test
JOB B test
JOB C test

# Pre script that fails for all nodes
SCRIPT PRE A {PYTHON3} {exe} 1
SCRIPT PRE B {PYTHON3} {exe} 1
SCRIPT PRE C {PYTHON3} {exe} 1

# Post script that fails for all nodes
SCRIPT POST A {PYTHON3} {exe} 1
SCRIPT POST B {PYTHON3} {exe} 1
SCRIPT POST C {PYTHON3} {exe} 1

# This should override all pre/post scripts to succeed
SCRIPT PRE ALL_NODES {PYTHON3} {exe} 0
SCRIPT POST ALL_NODES {PYTHON3} {exe} 0
""")

#-----------------------------------------------------------------------------------------
def setup_all_nodes_pre_skip(path: Path) -> str:
    exe = script_simple_exit(path)
    return write_dag_file(f"""
SUBMIT-DESCRIPTION test @=desc
    executable = {PYTHON3}
    arguments  = {exe} 0
    transfer_executable = false
    queue 1
@desc

JOB A test
JOB B test
JOB C test

# Pre script that fails for all nodes
SCRIPT PRE A {PYTHON3} {exe} 10
SCRIPT PRE B {PYTHON3} {exe} 10
SCRIPT PRE C {PYTHON3} {exe} 10

# Inform DAG to skip node if script exit 10
PRE_SKIP ALL_NODES 10
""")

#-----------------------------------------------------------------------------------------
def setup_all_nodes_retry(path: Path) -> str:
    script = script_check_retry(path)
    exe = script_simple_exit(path)
    return write_dag_file(f"""
SUBMIT-DESCRIPTION test @=desc
    executable = {PYTHON3}
    arguments  = {exe} 0
    transfer_executable = false
    queue 1
@desc

JOB A test
JOB B test
JOB C test

# Add post script to check set retry value
SCRIPT POST ALL_NODES {PYTHON3} {script} 10 $MAX_RETRIES

# Add retry value that will fail post script always
RETRY A 2 UNLESS-EXIT 125
RETRY B 2 UNLESS-EXIT 125
RETRY C 2 UNLESS-EXIT 125

# Override retry value to pass post script
RETRY ALL_NODES 10 UNLESS-EXIT 125
""")

#-----------------------------------------------------------------------------------------
def setup_all_nodes_abort_dag_on(path: Path) -> str:
    exe = script_simple_exit(path)
    return write_dag_file(f"""
JOB A @=desc
    executable = {PYTHON3}
    arguments  = {exe} 42
    transfer_executable = false
@desc

# Note can't easily check multiple node settings
ABORT-DAG-ON ALL_NODES 42 RETURN 2
""")

#-----------------------------------------------------------------------------------------
def setup_all_nodes_priority(path: Path) -> str:
    exe = script_simple_exit(path)
    return write_dag_file(f"""
SUBMIT-DESCRIPTION test @=desc
    executable = {PYTHON3}
    arguments  = {exe} 0
    transfer_executable = false
@desc

JOB A test
JOB B test
JOB C test

# Node priorities that will fail test
PRIORITY A 25
PRIORITY B 25
PRIORITY C 50

# Override node priorities with passing value
PRIORITY ALL_NODES 100
""")

def setup_all_nodes_category(path: Path) -> str:
    exe = script_simple_exit(path)
    return write_dag_file(f"""
SUBMIT-DESCRIPTION test @=desc
    executable = {PYTHON3}
    arguments  = {exe} 0
    transfer_executable = false
@desc

JOB A test
JOB B test
JOB C test

# Add all nodes into ALL category (expect 3 nodes in check)
CATEGORY ALL_NODES ALL
""")


#-----------------------------------------------------------------------------------------
def setup_all_nodes_splice(path: Path) -> str:
    exe = script_simple_exit(path)

    include = write_dag_file(f"""
SUBMIT-DESCRIPTION test @=desc
    executable = {PYTHON3}
    arguments  = {exe} $(code)
    transfer_executable = false
@desc
""", "job-desc.inc")

    splice = write_dag_file(f"""
INCLUDE {include}

JOB A test

# Ensure ALL_NODES stays within this splice
VARS ALL_NODES code=99
ABORT-DAG-ON ALL_NODES 99 RETURN 2
CATEGORY ALL_NODES SPLICED
""", "splice.dag")

    return write_dag_file(f"""
INCLUDE {include}

JOB A test
VARS A code=0

SPLICE S {splice}
""")

#-----------------------------------------------------------------------------------------
def verify_prio(condor, test_case, dag_handle) -> dict:
    schedd = condor.get_local_schedd()

    attempts = 0
    history = []

    while len(history) != 3:
        if attempts > 0:
            sleep(2)

        attempts += 1
        assert attempts <= 10, f"ERROR: Failed to get history all expected job records for test case {test_case}"
        history = schedd.history(
            constraint=f'DAGManJobId=={dag_handle.clusterid}',
            projection=["JobPrio"],
        )

    results = dict()
    for ad in history:
        prio = ad.get("JobPrio")
        if prio in results:
            results[prio] += 1
        else:
            results.update({prio : 1})

    return results

#-----------------------------------------------------------------------------------------
def verify_categories(condor, test_case, dag_handle) -> dict:
    categories = dict()

    KEY = "CHECK CATEGORY"

    log = DaemonLog(Path(f"{DAG_FILENAME}.dagman.out"))
    stream = log.open()
    for details in stream.read():
        if "D_TEST" in details.tags:
            line = str(details)
            # Line: <timestamp> (D_TEST) CHECK CATEGORY TotalNodes ThrottleLimit
            cat, num_nodes, _ = line[line.find(KEY) + len(KEY):].split()
            assert cat not in categories
            categories[cat] = int(num_nodes)

    return categories

#-----------------------------------------------------------------------------------------
class Details():
    def __init__(self, prep: Callable[[Path], str], ec: int, verifier: Tuple[Callable[[Condor, str, Handle], Any], Any] = None) -> None:
        self.prepare = prep
        self.exit_code = ec
        self.verification = verifier

TEST_CASES = {
    "VARS": Details(setup_all_nodes_vars, 0),
    "SCRIPTS": Details(setup_all_nodes_scripts, 0),
    "PRE_SKIP": Details(setup_all_nodes_pre_skip, 0),
    "RETRY": Details(setup_all_nodes_retry, 0),
    "ABORT_DAG_ON": Details(setup_all_nodes_abort_dag_on, 2),
    "PRIORITY": Details(setup_all_nodes_priority, 0, (verify_prio, {100:3})),
    "CATEGORY": Details(setup_all_nodes_category, 0, (verify_categories, {"ALL":3})),
    "SPLICE": Details(setup_all_nodes_splice, 2, (verify_categories, {"S+SPLICED":1})),
    "IGNORE_SPECIAL_NODES": Details(setup_special_nodes, 0),
}

#-----------------------------------------------------------------------------------------
@action
def run_dags(default_condor, test_dir):
    assert PYTHON3 is not None

    TESTS = dict()

    for TEST, SETUP in TEST_CASES.items():
        case_dir = test_dir / TEST
        if case_dir.exists():
            rmtree(case_dir)

        os.mkdir(case_dir)

        with ChangeDir(TEST):
            DAG = htcondor2.Submit.from_dag(SETUP.prepare(case_dir))

            HANDLE = default_condor.submit(DAG)
            TESTS[TEST] = (HANDLE, SETUP.exit_code, SETUP.verification)

    yield TESTS

#-----------------------------------------------------------------------------------------
@action(params={name: name for name in TEST_CASES})
def test_info(request, run_dags):
    TEST = run_dags[request.param]

    return (
        request.param,  # Test case name
        TEST[0],        # Test case job handle
        TEST[1],        # Test case expected DAG exit code
        TEST[2],        # Test case specific verifier
    )

@action
def test_case(test_info):
    return test_info[0]

@action
def test_dag_handle(test_info):
    HANDLE = test_info[1]
    assert HANDLE.wait(condition=ClusterState.all_complete, timeout=TIMEOUT)
    return HANDLE

@action
def test_expected_dag_ec(test_info, test_dag_handle):
    return test_info[2]

@action
def test_verification(test_info):
    verifier = test_info[3]
    if verifier is None:
        pytest.skip("Test case does not have special verification needed")

    return verifier

@action
def test_verification_result(default_condor, test_case, test_dag_handle, test_verification):
    with ChangeDir(test_case):
        return test_verification[0](default_condor, test_case, test_dag_handle)

@action
def test_verification_expected(test_verification):
    return test_verification[1]

#=========================================================================================
class TestDagAllNodes():
    def test_dag_exited_correctly(self, test_case, test_expected_dag_ec):
        with ChangeDir(test_case):
            with open(f"{DAG_FILENAME}.metrics", "r") as f:
                metrics = json.load(f)
                assert metrics["exitcode"] == test_expected_dag_ec

    def test_specific_verification(self, test_verification_result, test_verification_expected):
        assert test_verification_expected == test_verification_result




#!/usr/bin/env pytest

# Test user specifying a rescue file by filename

from ornithology import *
import htcondor2
import os
from pathlib import Path
from typing import Callable, Optional, List
from shutil import rmtree
import json
import glob

#------------------------------------------------------------------
TEST_FILE_EXT = "testfile"
TEST_SUBDAG_FILENAME = "sub.dag"
TEST_DAG_FILENAME = "test.dag"
CUSTOM_NAMED_RESCUE = "custom.rescue"

#------------------------------------------------------------------
# A script to touch a file using call {passed arg}.{extension}
@action
def the_job_script(test_dir):
    path = test_dir / "named_file.py"
    with open(path, "w") as f:
        f.write(f"""#!/usr/bin/env python3

import sys
from pathlib import Path

if len(sys.argv) != 2:
    print("ERROR: Script only expects 1 argument (unique filename modifier)", file=sys.stderr)
    args = " ".join(sys.argv[1:])
    print(f"       Args: '{{args}}'", file=sys.stderr)
    sys.exit(1)

p = Path(f"{{sys.argv[1]}}.{TEST_FILE_EXT}")
p.touch()
""")
    os.chmod(path, 0o755)
    return path

#------------------------------------------------------------------
# Create simple DAG with faux rescue files
def setup_simple_dag(script: Path) -> Path:
    dag_file = Path(TEST_DAG_FILENAME)
    with open(str(dag_file), "w") as f:
        f.write(f"""
SUBMIT-DESCRIPTION touch @=desc
    executable = {script}
    arguments  = "$(dag_node_name)"
    error      = job.debug
    output     = job.debug
    universe   = local
@desc

JOB A touch
JOB B touch
JOB C touch
""")

    for i in range(1, 5):
        rescue = f"{dag_file}.rescue{i:03d}"
        with open(rescue, "w") as f:
            f.write("DONE A\nDONE B\n")

    with open(CUSTOM_NAMED_RESCUE, "w") as f:
        f.write("DONE A\n")

    return dag_file

#------------------------------------------------------------------
# Create simple DAG and SUBDAG with faux rescue files
def setup_subdags(script: Path) -> Path:
    root_dag = Path(TEST_DAG_FILENAME)
    with open(str(root_dag), "w") as f:
        f.write(f"""
SUBMIT-DESCRIPTION touch @=desc
    executable = {script}
    arguments  = "$(dag_node_name)"
    error      = job.debug
    output     = job.debug
    universe   = local
@desc

JOB A touch
JOB B touch
JOB C touch

SUBDAG EXTERNAL D {TEST_SUBDAG_FILENAME}
""")

    with open(TEST_SUBDAG_FILENAME, "w") as f:
        f.write(f"""
SUBMIT-DESCRIPTION touch @=desc
    executable = {script}
    arguments  = "$(dag_node_name)"
    error      = job.debug
    output     = job.debug
    universe   = local
@desc

JOB AA touch
JOB BB touch
JOB CC touch
""")

    for i in range(1, 5):
        rescue = f"{TEST_DAG_FILENAME}.rescue{i:03d}"
        with open(rescue, "w") as f:
            f.write("DONE A\nDONE B\n")

    for i in range(1, 5):
        rescue = f"{TEST_SUBDAG_FILENAME}.rescue{i:03d}"
        with open(rescue, "w") as f:
            f.write("DONE BB\nDONE CC\n")

    with open(CUSTOM_NAMED_RESCUE, "w") as f:
        f.write("DONE A\n")

    return root_dag

#------------------------------------------------------------------
# Test case details
#     rescue: Rescue file to specify
#     rescue_num: Expected rescue number to found in root DAG metrics file
#     setup: Function to setup test case (i.e. write DAG & rescue files)
#     rescue_files: List of expected rescue files to find in test case directory
#     touch_files: List of files we expect from re-running DAGs (touched via the_job_script action script)
class Details():
    def __init__(self, rescue: str, rescue_num: int, setup: Callable[[Path], Path], rescue_files: List[str], touch_files: List[str]):
        self.rescue = rescue
        self.rescue_num = rescue_num
        self.setup = setup
        self.check_rescue_files = rescue_files
        self.check_touch_files = touch_files

#------------------------------------------------------------------
# Generator Class to produce expected rescue file names
#      base: Base DAG filename
#      ceiling: The highest rescue file number (non-inclusive)
#      old: Rescue file number to start marking files as old (inclusive)
class RescueGenerator:
    def __init__(self, base: str, ceiling: int, old: Optional[int] = None):
        self.base = base
        self.ceiling = ceiling
        self.old = old
        self.curr = 1

    def __iter__(self):
        return self

    def __next__(self):
        if self.curr < self.ceiling:
            num = self.curr
            self.curr += 1
            rescue = f"{self.base}.rescue{num:03d}"
            if self.old is not None and num >= self.old:
                rescue += ".old"
            return rescue
        raise StopIteration

#------------------------------------------------------------------
# Generator function to make all expected rescue files based on generator
# objects plus the custom named rescue file
def gen_expected_rescues(generators: List[RescueGenerator]):
    yield CUSTOM_NAMED_RESCUE
    for gen in generators:
        for rescue in gen:
            yield rescue

#------------------------------------------------------------------
TEST_CASES = {
    "RESCUE_MOST_RECENT": Details(
        f"{TEST_DAG_FILENAME}.rescue004",
        4,
        setup_simple_dag,
        list(gen_expected_rescues([RescueGenerator(TEST_DAG_FILENAME, 5)])),
        [f"C.{TEST_FILE_EXT}"]
    ),
    "RESCUE_SPECIFIC": Details(
        f"{TEST_DAG_FILENAME}.rescue002",
        2,
        setup_simple_dag,
        list(gen_expected_rescues([RescueGenerator(TEST_DAG_FILENAME, 5, 3)])),
        [f"C.{TEST_FILE_EXT}"]
    ),
    "RESCUE_CUSTOM": Details(
        CUSTOM_NAMED_RESCUE,
        0,
        setup_simple_dag,
        list(gen_expected_rescues([RescueGenerator(TEST_DAG_FILENAME, 5, 1)])),
        [f"B.{TEST_FILE_EXT}", f"C.{TEST_FILE_EXT}"]
    ),
    "RESCUE_SUBDAG_SPECIFIC": Details(
        f"{TEST_DAG_FILENAME}.rescue001",
        1,
        setup_subdags,
        list(gen_expected_rescues([RescueGenerator(TEST_DAG_FILENAME, 5, 2), RescueGenerator(TEST_SUBDAG_FILENAME, 5, 2)])),
        [f"C.{TEST_FILE_EXT}", f"AA.{TEST_FILE_EXT}"],
    ),
    "RESCUE_SUBDAG_CUSTOM": Details(
        CUSTOM_NAMED_RESCUE,
        0,
        setup_subdags,
        list(gen_expected_rescues([RescueGenerator(TEST_DAG_FILENAME, 5, 1), RescueGenerator(TEST_SUBDAG_FILENAME, 5)])),
        [f"B.{TEST_FILE_EXT}", f"C.{TEST_FILE_EXT}", f"AA.{TEST_FILE_EXT}"],
    ),
}

#==================================================================
# Submit all the DAGS in one loop
@action
def run_dags(default_condor, test_dir, the_job_script):
    RESULTS = dict()

    for CASE, DETAILS in TEST_CASES.items():
        exe_dir = Path(CASE)
        if exe_dir.exists():
            rmtree(exe_dir)
        exe_dir.mkdir()

        with ChangeDir(exe_dir):
            dag_file = DETAILS.setup(the_job_script)

            with default_condor.use_config():
                DAG = htcondor2.Submit().from_dag(str(dag_file), {"RescueFile" : DETAILS.rescue})

            HANDLE = default_condor.submit(DAG)
            RESULTS[CASE] = (HANDLE, DETAILS.check_rescue_files, DETAILS.check_touch_files, DETAILS.rescue_num)

    yield RESULTS

#------------------------------------------------------------------
@action(params={name: name for name in TEST_CASES.keys()})
def test_info(request, run_dags):
    return (request.param, run_dags[request.param][0], run_dags[request.param][1], run_dags[request.param][2], run_dags[request.param][3])

@action
def the_test_case(test_info):
    return test_info[0]

@action
def the_test_handle(test_info):
    return test_info[1]

@action
def the_expected_rescue_files(test_info):
    return test_info[2]

@action
def the_expected_touch_files(test_info):
    return test_info[3]

@action
def the_expected_rescue_num(test_info):
    return test_info[4]

#==================================================================
class FailTestException(Exception):
    """Custom exception to raise for test case check failure"""
    pass

class PassTestException(Exception):
    """Custom exception to raise for test case check passing condition"""
    pass

#------------------------------------------------------------------
class TestDAGManSpecifiedRescue:
    def test_dag_execute_in_rescue_mode(self, the_test_case, the_test_handle, the_expected_rescue_num):
        with ChangeDir(the_test_case):
            # Wait for root DAG handle to finish
            assert the_test_handle.wait(condition=ClusterState.all_complete, timeout=120)

            # Check for root DAG metrics (completed successfully and rescue num)
            with open(f"{TEST_DAG_FILENAME}.metrics", "r") as f:
                stats = json.load(f)
                assert stats["exitcode"] == 0
                assert stats["rescue_dag_number"] == the_expected_rescue_num

            # Check root DAG debug file for specific rescue file debug message
            with open(f"{TEST_DAG_FILENAME}.dagman.out", "r") as f:
                try:
                    for line in f:
                        if "Rescue DAG file specified" in line:
                            raise PassTestException
                    raise FailTestException
                except Exception as e:
                    assert isinstance(e, PassTestException)

    def test_check_rescue_files(self, the_test_case, the_expected_rescue_files):
        located_rescues = list()

        # Check all files with *.rescue* in test case directory are expected
        with ChangeDir(the_test_case):
            for rescue in glob.glob("*.rescue*"):
                assert os.path.isfile(rescue)
                assert rescue in the_expected_rescue_files
                located_rescues.append(rescue)

        # Check we found all the expected rescue files
        located_rescues.sort()
        the_expected_rescue_files.sort()
        assert located_rescues == the_expected_rescue_files

    def test_check_touch_files(self, the_test_case, the_expected_touch_files):
        located_touch_files = list()

        # Check that all *.{extension} files in test case directory are expected
        with ChangeDir(the_test_case):
            for rescue in glob.glob(f"*.{TEST_FILE_EXT}"):
                assert os.path.isfile(rescue)
                assert rescue in the_expected_touch_files
                located_touch_files.append(rescue)

        # Check we found all expected touched files
        located_touch_files.sort()
        the_expected_touch_files.sort()
        assert located_touch_files == the_expected_touch_files

